/*
 * Teletronics WL24xx PCMCIA driver from WebWorld Warehouse Ltd.
 * Released May 30, 2000
 *
 * This driver is based on ZCom's WL2420 FreeBSD 2.2.5-RELEASE source.
 *
 * We have ported this driver to FreeBSD 3.4-RELEASE, and has been tested
 * in both SMP and non-SMP modes, using firmware 2.0.3 and 2.0.6. 
 *
 * This driver is not pretty, but it does work. We have fixed many 
 * problems that the original driver had, and now performs quite well.
 *
 * Device name has changed from zwl to twl, as to not conflict with 
 * future ZCom driver releases.  
 *
 * Contact: wireless@valemount.com
 * Website: www.webworldwarehouse.com
 */

/******************************************************************************
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *
 *****************************************************************************/

/*
 * WL24xx Wireless LAN Card Driver for FreeBSD
 *
 * References:
 *   1. WL24xx packet drivers (tooasm.asm)
 *   2. Access Point Firmware Interface Specification for IEEE 802.11 SUTRO
 *   3. IEEE 802.11
 *   4. Linux network driver (/usr/src/linux/drivers/net)
 *   5. WL24xx driver for Linux
 *   6. 3c505 FreeBSD driver (ISA:if_eg, PCMCIA:if_ed, if_ep)
 *
 * The structure of FreeBSD network drivers is more confused than Linux.
 * It doesn't split upper half from bottom layer. You will see lots of
 * threads mixed together.
 *
 * To catch the idea, I illustrate all program flows as follows:
 *
 * ISA bus:
 *
 *   Thread 1: zwl_probe_isa() -> zwl_attach() -> zwl_ioctl() -> zwl_init() -> zwl_start()
 *   Thread 2: zwlintr() --> zwl_rxintr() -> zwl_read() -> zwl_get()
 *                       \-> zwl_start()
 *   Thread 3: zwl_ioctl() -> zwl_setfilter()
 *   Thread 4: zwl_watchdog() -> zwl_reset()
 *
 * PCCARD (PCMCIA) bus:
 *
 *   Thread 1: zwl_init_pccard() -> zwl_probe_pccard() -> zwl_attach -> ...
 *   Thread 2: zwl_intr_pccard() -> zwlintr() -> ...
 *   Thread 3: zwl_unload_pccard() -> if_down()
 *   Reset refer ISA
 *
 * Tested with WL2420 firmware 1.2, FreeBSD 2.2.5-RELEASE
 */


#include "bpfilter.h"
#include "card.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/sockio.h>
#include <sys/mbuf.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/socket.h>
#include <sys/syslog.h>
#include <sys/ucred.h>
#include <sys/errno.h>
#include <sys/proc.h>

#include <net/if.h>
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <net/if_dl.h>
#include <net/if_media.h>
#include <net/if_types.h>

#ifdef INET
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/if_ether.h>
#endif

#if NBPFILTER > 0
#include <net/bpf.h>
#endif

#include <machine/clock.h>
#include <machine/md_var.h>
#include <machine/bus_pio.h>
#include <machine/bus.h>

#include <i386/isa/isa_device.h>
#include <i386/isa/icu.h>
#include <i386/isa/if_wireg.h>
#include <machine/if_wavelan_ieee.h>

#if NCARD > 0
#include <sys/select.h>
#include <pccard/cardinfo.h>
#include <pccard/slot.h>
#endif

/* Enable Debug */
/*#define WLAPI_DEBUG*/


#ifndef __WLAPI_H__
#define __WLAPI_H__

#ifndef __FreeBSD__
#define __FreeBSD__
#endif

/*#define LINUX_DRIVER*/
/* Uncomment the line, if you would like to debug API */
/*#define WLAPI_DEBUG*/

/* Uncomment the line, if the driver runs unstablely on notebook */
#define WL_SLOWDOWN_IO
#ifdef __FreeBSD__
/*-------------------------- FreeBSD Begin ----------------------------*/
#  define  FREEBSD_DRIVER
/*#  define SLOW_DOWN_IO __asm__ __volatile__("outb %al,$0x80")*/
#  define SLOW_DOWN_IO	DELAY(1)

   /* inb(), and outb() are defined within i386/include/cpufunc.h */
#  ifdef  WL_SLOWDOWN_IO
#	 define InB(port)				  inb(port)
#	 define OutB(data, port)		  { outb((port), (data)); SLOW_DOWN_IO; }
#	 define OutB_P(data, port)		  { outb((port), (data)); SLOW_DOWN_IO; SLOW_DOWN_IO; }
#	 define OutSB(port, addr, size)   { outsb((port), (addr), (size)); SLOW_DOWN_IO; }
#  else
#	 define InB(port)				  inb(port)
#	 define OutB(data, port)		  outb((port), (data));
#	 define OutB_P(data, port)		  { outb((port), (data)); SLOW_DOWN_IO; }
#	 define OutSB(port, addr, size)   outsb((port), (addr), (size))
#  endif /* !SLOW_DOWN_IO */
#  define printk  printf

   void* memset(void *pBuf, int nValue, size_t nSize);	/* my own */
/*---------------------------- FreeBSD End ----------------------------*/
#else
/*---------------------------- Linux Begin ----------------------------*/
#  define  LINUX_DRIVER
#  include <linux/netdevice.h>			/* for enet_statistics */
#  include <linux/kernel.h>
#  include <linux/string.h>
#  include <asm/io.h>
#  ifdef WL_PCMCIA
#	 include <pcmcia/version.h>
#	 include <pcmcia/cs_types.h>
#	 include <pcmcia/cs.h>
#	 include <pcmcia/cistpl.h>
#	 include <pcmcia/ds.h>
#  endif /* End of WL_PCMCIA */

#  define ETHER_MIN_LEN 			64
#  define ETHER_MAX_LEN 			1518

   /* inb(), and outb() are defined within include/asm/io.h */
#  ifdef  WL_SLOWDOWN_IO
#	 define InB(port)				  inb(port)
#	 define OutB(data, port)		  { outb((data), (port)); SLOW_DOWN_IO; }
#	 define OutB_P(data, port)		  { outb_p((data), (port)); SLOW_DOWN_IO; }
#	 define OutSB(port, addr, size)   { outsb((port), (addr), (size)); SLOW_DOWN_IO; }
#  else
#	 define InB(port)				  inb(port)
#	 define OutB(data, port)		  outb((data), (port))
#	 define OutB_P(data, port)		  outb_p((data), (port))
#	 define OutSB(port, addr, size)   outsb((port), (addr), (size))
#  endif
/*---------------------------- Linux End   ----------------------------*/
#endif

#ifdef __alpha__
typedef unsigned int			DWORD;
#else
typedef unsigned long		DWORD;
#endif

typedef unsigned short			WORD;
typedef unsigned char			BYTE;
typedef int 					BOOL;

/* define for WLA 2.0 */

#define USHORT WORD
#define UCHAR  unsigned char
#define BlockSize 256
/***************************************************************
 ID for input Signals of DRIVER block
 bit[7-5] is block ID: 000
 bit[4-0] is signal ID
 ***************************************************************/
#define Alarm				0x00
#define MdConfirm			0x01
#define MdIndicate			0x02
#define AssocConfirm		0x03
#define AssocIndicate		0x04
#define AutheConfirm		0x05
#define AutheIndicate		0x06
#define DeautheConfirm		0x07
#define DeautheIndicate 	0x08
#define DisassocConfirm 	0x09
#define DisassocIndicate	0x0A
#define GetConfirm			0x0B
#define JoinConfirm 		0x0C
#define PowermgtConfirm 	0x0D
#define ReassocConfirm		0x0E
#define ReassocIndicate 	0x0F
#define ScanConfirm 		0x10
#define SetConfirm			0x11
#define StartConfirm		0x12
#define ResyncConfirm		0x13
#define SiteConfirm 		0x14
#define SaveConfirm 		0x15
#define RFtestConfirm		0x16

/***************************************************************
 ID for input Signals of MLME block
 bit[7-5] is block ID: 010
 bit[4-0] is signal ID
 ***************************************************************/
#define AssocRequest	   0x20
#define AutheRequest	   0x21
#define DeautheRequest	   0x22
#define DisassocRequest    0x23
#define GetRequest		   0x24
#define JoinRequest 	   0x25
#define PowermgtRequest    0x26
#define ReassocRequest	   0x27
#define ScanRequest 	   0x28
#define SetRequest		   0x29
#define StartRequest	   0x2A
#define MdRequest		   0x2B
#define ResyncRequest	   0x2C
#define SiteRequest 	   0x2D
#define SaveRequest 	   0x2E
#define RFtestRequest	   0x2F

#define MmConfirm		   0x60
#define MmIndicate		   0x61

#define Infrastructure	   0
#define Independent 	   1
#define Any_bss 		   2
#define ActiveScan		   0
#define PassiveScan 	   1
#define TxResult_Success	0
#define TxResult_NoBss		1
#define TxResult_retryLimit 2

#define Open_System 		0
#define Share_Key			1

#define EtherII 0
#define Ether802_3e 1
#define Ether802_3f 2

#define SUCCESS 	0
#define INVALID 	1
#define TIMEOUT 	2
#define REFUSED 	3
#define MANY_REQ	4
#define ALREADYBSS	5

#define UNSPEC		1
#define PRE_AU_NOV	2
#define DEAUTH3 	3
#define DIASS2INA	4
#define CLASS2		5
#define CLASS3		6
#define DIASS8		7
#define NOAUTH9 	8
#define RESERVED10	9
/* MIB Attrib */
#define aStationID							  0
#define aAuthenticationAlogrithms			  1
#define aAuthenticationType 				  2
#define aMediumOccupancyLimit				  3
#define aCFPollable 						  4
#define aCFPPeriod							  5
#define aCFPMaxDuration 					  6
#define aAuthenticationResponseTimeout		  7
#define aReceiveDTIMs						  8
#define aPrivacyOptionImplemented			  9
#define aPrivacyInvoked 					 10
#define aWEPDefaultKeys 					 11
#define aWEPDefaultKeyID					 12
#define aWEPKeyMappings 					 13
#define aWEPKeyMappingLength				 14
#define aExcludeUnencrypted 				 15
#define aWEPICVErrorCount					 16
#define aWEPUndecryptableCount				 17
#define aWEPExcludedCount					 18
#define aMacAddress 						 19
#define aGroup_Addresses					 20
#define aRTSThreshold						 21
#define aShortRetryLimit					 22
#define aLongRetryLimit 					 23
#define aFragmentationThreshold 			 24
#define aMaxTransmitMSDULifetime			 25
#define aMaxReceiveLifetime 				 26
#define aManufacturerID 					 27
#define aProductID							 28
#define aTransmittedFragmentCount			 29
#define aMulticastTransmittedFrameCount 	 30
#define aFailedCount						 31
#define aReceiveFragmentCount				 32
#define aMulticastReceivedCount 			 33
#define aFCSErrorCount						 34
#define aRetryCount 						 35
#define aMultipleRetryCount 				 36
#define aRTSSuccessCount					 37
#define aRTSFailureCount					 38
#define aACKFailureCount					 39
#define aFrameDuplicateCount				 40
#define aPHYType							 41
#define aRegDomainsSupport					 42
#define aCurrentRegDomain					 43
#define aSlotTime							 44
#define aCCATime							 45
#define aRxTxTurnaroundTime 				 46
#define aTxPLCPDelay						 47
#define aRxTxSwitchTime 					 48
#define aTxRampOnTime						 49
#define aTxRFDelay							 50
#define aSIFSTime							 51
#define aRxRFDelay							 52
#define aRxPLCPDelay						 53
#define aMACProcessingDelay 				 54
#define aTxRampOffTime						 55
#define aPreambleLength 					 56
#define aPLCPHeaderLength					 57
#define aMPDUDurationFactor 				 58
#define aAirPropagationTime 				 59
#define aTempType							 60
#define aCWmin								 61
#define aCWmax								 62
#define aSupportDataRatesTx 				 63
#define aSupportDataRatesRx 				 64
#define aMPDUMaxLength						 65
#define aSupportTxAntennas					 66
#define aCurrentTxAntenna					 67
#define aSupportRxAntennas					 68
#define aDiversitySupport					 69
#define aDiversitySelectionRx				 70
#define aNumberSupportedPowerLevels 		 71
#define aTxPowerLevel1						 72
#define aTxPowerLevel2						 73
#define aTxPowerLevel3						 74
#define aTxPowerLevel4						 75
#define aTxPowerLevel5						 76
#define aTxPowerLevel6						 77
#define aTxPowerLevel7						 78
#define aTxPowerLevel8						 79
#define aCurrentTxPowerLevel				 80
#define aCurrentChannel 					 81
#define aCCAModeSupported					 82
#define aCurrentCCAMode 					 83
#define aEDThreshold						 84
#define aSynthesizerLocked					 85
#define aCurrentPowerState					 86
#define aDozeTurnonTime 					 87
#define aRCR33								 88
#define aDefaultChannel 					 89
#define aSSID								 90
#define aPowerMgmtEnable					 91
#define aNetworkCapability					 92
#define aRouting							 93



typedef struct t_Segment256{
	USHORT	NextSegment;
	UCHAR	 data[254];
} Segment256;
typedef struct t_Segment32{
	USHORT	NextSegment;
	UCHAR	 data[30];
} Segment32;
typedef struct t_TxHeader{
	USHORT	XmitCount;
	UCHAR	SYNC[16];
	USHORT	SFD;
	UCHAR	SIGNAL;
	UCHAR	SERVICE;
	USHORT	LENGTH;
	USHORT	CRC16;
	USHORT	FrameControl;
	USHORT	DurationID;
	UCHAR	Address1[6];
	UCHAR	Address2[6];
	UCHAR	Address3[6];
	USHORT	SequenceControl;
	UCHAR	Address4[6];
} TxHeader;
typedef struct t_RxHeader{
	USHORT	RxNextBlock;
	USHORT	RxNextFrameBlock;
	UCHAR	RxBlockCtrl;
	UCHAR	RxNextFrame;
	UCHAR	RxNextFrame1;
	UCHAR	RSSI;
	UCHAR	Time[8];
	UCHAR	Signal;
	UCHAR	Service;
	USHORT	Length;
	USHORT	CRC16;
	USHORT	FrameControl;
	USHORT	Duration;
	UCHAR	Address1[6];
	UCHAR	Address2[6];
	UCHAR	Address3[6];
	USHORT	Sequence;
	UCHAR	Address4[6];
}RxHeader;

typedef struct t_StartReq{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR BSSType;
	USHORT	BeaconPeriod;
	USHORT	DTIMPeriod;
	USHORT	ProbeDelay;
	USHORT	CapablilityInfo;
	UCHAR SSID[34];
	UCHAR BSSBasicRateSet[10];
	UCHAR OperationalRateSet[10];
	UCHAR CFpset[8];
	UCHAR PHYpset[3];
	UCHAR IBSSpset[4];
} StartReq;

typedef struct t_AssocReq{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Timeout;
	USHORT	CapInfo;
	USHORT	ListenInterval;
	UCHAR MacAddress[6];
} AssocReq;
typedef struct t_AssocConf{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Status;
} AssocConf;
typedef struct t_AssocInd{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR MacAddress[6];
} AssocInd;
typedef struct t_AutheReq{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Type;
	USHORT	Timeout;
	UCHAR MacAddress[6];
} AutheReq;
typedef struct t_AutheConf{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Type;
	USHORT	Status;
	UCHAR  MacAddress[6];
} AutheConf;
typedef struct t_AutheInd{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Type;
	UCHAR MacAddress[6];
} AutheInd;
typedef struct t_DeautheReq{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Reason;
	UCHAR  MacAddress[6];
} DeautheReq;
typedef struct t_DeautheConf{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Status;
	UCHAR  MacAddress[6];
} DeautheConf;
typedef struct t_DeautheInd{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Reason;
	UCHAR  MacAddress[6];
} DeautheInd;
typedef struct t_DisassocReq{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Reason;
	UCHAR  MacAddress[6];
} DisassocReq;
typedef struct t_DisassocConf{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Status;
} DisassocConf;
typedef struct t_DisassocInd{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Reason;
	UCHAR  MacAddress[6];
} DisassocInd;
typedef struct t_GetReq{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	MibAtrib;
} GetReq;
typedef struct t_GetConf{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	MibStatus;
	USHORT	MibAtrib;
	UCHAR	MibValue[100];
} GetConf;
typedef struct t_JoinReq{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	UCHAR OperationalRateSet[10];
	USHORT	Reserved2;
	USHORT	Timeout;
	USHORT	ProbeDelay;
	UCHAR Timestamp[8];
	UCHAR LocalTime[8];
	USHORT	BeaconPeriod;
	USHORT	DTIMPeriod;
	USHORT	CapablilityInfo;
	UCHAR BSSType;
	UCHAR  BSSID[6];
	UCHAR SSID[34];
	UCHAR PHYpset[3];
	UCHAR CFpset[8];
	UCHAR IBSSpset[4];
	UCHAR BSSBasicRateSet[10];
} JoinReq;
typedef struct t_JoinConf{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Status;
} JoinConf;

typedef struct t_PowerMgtReq{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR PwrSave;
	UCHAR WakeUp;
	UCHAR ReceiveDTIMs;
} PowerMgtReq;
typedef struct t_PowerMgtConf{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Status;
} PowerMgtConf;

typedef struct t_ReassocReq{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Timeout;
	USHORT	CapInfo;
	USHORT	ListInt;
	UCHAR  MacAddress[6];
} ReassocReq;
typedef struct t_ReassocConf{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Status;
} ReassocConf;
typedef struct t_ReassocInd{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	UCHAR  MacAddress[6];
} ReassocInd;
typedef struct t_ResetReq{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	UCHAR SetMIB;
	UCHAR  MacAddress[6];
} ResetReq;
typedef struct t_ResetConf{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Status;
} ResetConf;
typedef struct t_ScanReq{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR BSSType;
	USHORT	ProbeDelay;
	USHORT	MinChannelTime;
	USHORT	MaxChannelTime;
	UCHAR ChannelList[14];
	UCHAR  BSSID[6];
	UCHAR SSID[34];
	UCHAR ScanType;
} ScanReq;
typedef struct t_ScanConf{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Status;
	UCHAR Timestamp[8];
	UCHAR LocalTime[8];
	USHORT	BeaconPeriod;
	USHORT	DTIMPeriod;
	USHORT	CapablilityInfo;
	UCHAR BSSType;
	UCHAR BSSID[6];
	UCHAR SSID[34];
	UCHAR PHYpset[3];
	UCHAR CFpset[8];
	UCHAR IBSSpset[4];
	UCHAR BSSBasicRateSet[10];
	UCHAR RSSI;
} ScanConf;

typedef struct t_SetReq{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	MibAtrib;
	USHORT	MibValue;
} SetReq;
typedef struct t_SetConf{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	MibStatus;
	USHORT	MibAtrib;
} SetConf;
typedef struct t_StartConf{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Status;
} StartConf;
typedef struct t_MdReq{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Routing;
	USHORT	Data;
	USHORT	Size;
	UCHAR Priority;
	UCHAR ServiceClass;
	UCHAR DAddress[6];
	UCHAR SAddress[6];
} MdReq;
typedef struct t_MdInd{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Routing;
	USHORT	Data;
	USHORT	Size;
	UCHAR Reception;
	UCHAR Priority;
	UCHAR ServiceClass;
	UCHAR DAddress[6];
	UCHAR SAddress[6];
} MdInd;
typedef struct t_MdCon{
	USHORT	NextBlock;
	UCHAR SignalID;
	UCHAR Reserved;
	USHORT	Data;
	UCHAR Status;
	UCHAR Priority;
	UCHAR ServiceClass;
} MdCon;

typedef struct t_ResyncReq{
	USHORT	NextBlock;
	UCHAR SignalID;
} ResyncReq;

typedef struct t_ResyncConf{
	USHORT	NextBlock;
	UCHAR	SignalID;
	UCHAR	Status;
} ResyncConf;

typedef struct t_SiteReq{
	USHORT	NextBlock;
	UCHAR SignalID;
} SiteReq;
typedef struct t_SiteConf{
	USHORT	NextBlock;
	UCHAR	SignalID;
	UCHAR	Reserved;
	USHORT	Status;
	UCHAR	RSSI[101];
} SiteConf;
typedef struct t_SaveReq{
	USHORT	NextBlock;
	UCHAR	SignalID;
} SaveReq;
typedef struct t_SaveConf{
	USHORT	NextBlock;
	UCHAR	SignalID;
	UCHAR	Status;
} SaveConf;


/*----------------------- WLAPI_DEBUG definition-----------------------*/
#ifdef	WLAPI_DEBUG
#  define WL_DBG_API			0x00000001L
#  define WL_DBG_INIT			0x00000002L
#  define WL_DBG_EXCEPT 		0x00000004L
#  define WL_DBG_ROAM			0x00000008L
#  define WL_DBG_PNP			0x00000100L

#  define WL_STAT_API_FUNC		0x000000FFL
#  define WL_STAT_INIT			0x00000001L
#  define WL_STAT_TX_PKT		0x00000002L
#  define WL_STAT_RX_LOOK		0x00000004L
#  define WL_STAT_RX_RECV		0x00000008L
#  define WL_STAT_RX_DONE		0x00000010L

#  define WL_STAT_LOCK_FUNC 	0x00000F00L
#  define WL_STAT_LOCK_SUTRO	0x00000100L
#  define WL_STAT_UNLOCK_SUTRO	0x00000200L

#  define WL_STAT_MIB_FUNC		0x0000F000L
#  define WL_STAT_GET_MIB		0x00001000L
#  define WL_STAT_SET_MIB		0x00002000L

#  define WL_STAT_TRIVIAL_FUNC	0x00FF0000L
#  define WL_STAT_BLK_INTR		0x00010000L
#  define WL_STAT_UNBLK_INTR	0x00020000L
#  define WL_STAT_INT_WLA		0x00040000L
#  define WL_STAT_SET_WLA		0x00080000L
#  define WL_STAT_GET_WLA		0x00100000L
#  define WL_STAT_FLASH_READ	0x00200000L
#  define WL_STAT_PNP_PROBE 	0x00400000L
#  define WL_STAT_PNP_SET		0x00800000L

   extern DWORD WL_DbFlag;
#  define TRACE_API(format, args...)  \
			{ if (WL_DbFlag & WL_DBG_API) printk(format , ## args); }
#  define TRACE_INIT(format, args...)  \
			{ if (WL_DbFlag & WL_DBG_INIT) printk(format , ## args); }
#  define TRACE_EXCEPT(format, args...)  \
			{ if (WL_DbFlag & WL_DBG_EXCEPT) printk(format , ## args); }
#  define TRACE_ROAM(format, args...)  \
			{ if (WL_DbFlag & WL_DBG_ROAM) printk(format , ## args); }
#  define TRACE_PNP(format, args...)  \
			{ if (WL_DbFlag & WL_DBG_PNP) printk(format , ## args); }
#  define TRACE(format, args...)	printk(format , ## args);
#  define ASSERT(x) \
			{ if (!(x)) printk("FILE:%s LINE:%d, ASSERT FAILED\n", __FILE__, __LINE__); }
#  define WHEN_WLAPI_DEBUG(x) x
#  define ASSERT_STAT(x) \
			{ (x) ? 0 : printk("FILE:%s LINE:%d, ASSERT FAILED! STAT=0x%lx\n", __FILE__, __LINE__, pAdaptor->ulDbgStat); }
#else
#  define TRACE_API(format, args...)
#  define TRACE_INIT(format, args...)
#  define TRACE_EXCEPT(format, args...)
#  define TRACE_ROAM(format, args...)
#  define TRACE_PNP(format, args...)
#  define TRACE(format, args...)
#  define ASSERT(x)
#  define WHEN_WLAPI_DEBUG(x)
#  define ASSERT_STAT(x)
#endif

/* For rough constant delay */

#define NOPLOOP(n)	   { int x=0; while(x++<n) SLOW_DOWN_IO; }
/*
#define NOPLOOP(n)	   DELAY(n)
*/
/**************************************************************************/
/* !!!!!!!		  All structures must be tight alignmented		  !!!!!!!!*/
/**************************************************************************/
#pragma pack (1)

/* Ethernet MAC Address, BSS_ID, or ESS_ID */
/* With this, we may simply write "x=y;" instead of "memcpy(x, y, 6);" */
/* It's more efficiency with compiler's optimization and more clearly  */
typedef struct _MADDR {
	BYTE	b0;
	BYTE	b1;
	BYTE	b2;
	BYTE	b3;
	BYTE	b4;
	BYTE	b5;
} MADDR;

/* Instead of calling memcmp(), use those macro to speed up! */
/* To port to platforms that care alignment, I do not do DWORD comparing */
#define MA_ISEQUAL(x, y)   ( (x).b0 == (y).b0 && (x).b1 == (y).b1 && (x).b2 == (y).b2 && (x).b3 == (y).b3 && (x).b4 == (y).b4 && (x).b5 == (y).b5 )

#define MA_ISZERO(x)	   ( (x).b0 == 0 && (x).b1 == 0 && (x).b2 == 0 && (x).b3 == 0 && (x).b4 == 0 && (x).b5 == 0 )

#ifndef TRUE
#  define  FALSE  0
#  define  TRUE   (!FALSE)
#endif

/* */
/* Definitions for supporting clone adapters. */
/* */

/* Max TX_DESC reserved */
#define MAX_TX_DESC_RESERVED	20

/* Timeout for beacon (unit: ms.) */
#define BEACON_TIMEOUT			1000

/* */
/* Valid value ranges for the IoBaseAddress. */
/* */
#define MIN_IOBASEADDR 0x0100
#define MAX_IOBASEADDR 0xffc0

#define WL_IO_REGION			0x10

/* */
/* Default value for Adapter->IoBaseAddr */
/* */
#define DEFAULT_IOBASEADDR (PVOID)0x280

/* */
/* Default value for Adapter->InterruptNumber */
/* */
#define DEFAULT_INTERRUPTNUMBER 11


/* System Interface Registers (SIR space) */
#define NIC_GCR 	((BYTE)0x00)   /* SIR0 - General Configuration Register */
#define NIC_BSS 	((BYTE)0x01)   /* SIR1 - Bank Switching Select Register */
#define NIC_LMAL	((BYTE)0x02)   /* SIR2 - Local Memory Address Register [7:0] */
#define NIC_LMAH	((BYTE)0x03)   /* SIR3 - Local Memory Address Register [14:8] */
#define NIC_IODPA	((BYTE)0x04)   /* SIR4 - I/O Data Port A */
#define NIC_IODPB	((BYTE)0x05)   /* SIR5 - I/O Data Port B */
#define NIC_IODPC	((BYTE)0x06)   /* SIR6 - I/O Data Port C */
#define NIC_IODPD	((BYTE)0x07)   /* SIR7 - I/O Data Port D */


/* Bits in GCR */
#define GCR_SWRESET 	((BYTE)0x80)
#define GCR_CORESET 	((BYTE)0x40)
#define GCR_DISPWDN 	((BYTE)0x20)
#define GCR_ECWAIT		((BYTE)0x10)
#define GCR_ECINT		((BYTE)0x08)
#define GCR_INT2EC		((BYTE)0x04)
#define GCR_ENECINT 	((BYTE)0x02)
#define GCR_DAM 		((BYTE)0x01)


/* Bits in BSS (Bank Switching Select Register) */
#define BSS_FPAGE0		((BYTE)0x20)   /* Flash memory page0 */
#define BSS_FPAGE1		((BYTE)0x28)
#define BSS_FPAGE2		((BYTE)0x30)
#define BSS_FPAGE3		((BYTE)0x38)
#define BSS_SPAGE0		((BYTE)0x00)   /* SRAM page0 */
#define BSS_SPAGE1		((BYTE)0x08)
#define BSS_SPAGE2		((BYTE)0x10)
#define BSS_SPAGE3		((BYTE)0x18)


/**************************************************************** */
/* Define Driver Interface */
/**************************************************************** */

/*
 * Refer IEEE 802.11
 */

/* */
/* Tx packet header, include PLCP and MPDU */
/* */

/*
 * Tx PLCP Header
 */
typedef struct _TX_PLCP_HDR {
	BYTE	SYNC[16];
	WORD	SFD;
	BYTE	SIGNAL;
	BYTE	SERVICE;
	WORD	LENGTH;
	WORD	CRC16;
} TX_PLCP_HDR, *PTX_PLCP_HDR;

/*
 * Data Frame MAC Header (IEEE 802.11)
 */
typedef struct _DATA_MAC_HDR {
	WORD	FrameControl;
	WORD	DurationID;
	MADDR	Address1;
	MADDR	Address2;
	MADDR	Address3;
	WORD	SequenceControl;
	MADDR	Address4;
} DATA_MAC_HDR, *PDATA_MAC_HDR;

typedef struct _TX_HEADER{
	TX_PLCP_HDR 	PlcpHdr;
	DATA_MAC_HDR	MacHdr;
} TX_HEADER, * PTX_HEADER;

#define FRAME_CTRL_INFRA	((WORD) 0x0108)
#define FRAME_CTRL_ADHOC	((WORD) 0x0008)
#define FRAME_CTRL_AP		((WORD) 0x0208)
#define FRAME_CTRL_IAP		((WORD) 0x0308)
#define FRAME_CTRL_BEACON_LO   ((BYTE) 0x80)

/* */
/* Result of Ne2000IndicatePacket(). */
/* */
typedef enum {
	INDICATE_OK,
	SKIPPED,
	ABORT,
	CARD_BAD
} INDICATE_STATUS;


/**************************************************************************/
/* !!!!!!!				Restore old alignment setup 			  !!!!!!!!*/
/**************************************************************************/
#pragma pack ()


/*

   Reserve the beginning Tx space for descriptor use.


   TxBlockOffset -->  *----*----*----*----* \
	(TxFreeDesc)	  |  0 |  1 |  2 |	3 |  \
					  |  4 |  5 |  6 |	7 |   |
					  |  8 |  9 | 10 | 11 |   TX_DESC * 20
					  | 12 | 13 | 14 | 15 |   |
					  | 16 | 17 | 18 | 19 |  /
   TxBufferBegin -->  *----*----*----*----* /
   (TxBufferHead)	  | 				  |
   (TxBufferTail)	  | 				  |
					  |    Send Buffer	  |
					  | 				  |
					  | 				  |
					  *-------------------*
   TxBufferEnd -------------------------/

*/

typedef struct _WL_Adaptor {
	int 	nBaseAddr;
	MADDR	MacAddress;
	WORD	uTxBufferSize;
	WORD	uTxBufferHead;
	WORD	uTxBufferTail;
	WORD	uTxBufferCount;
	WORD	uESBQReqStart;
	WORD	uESBQReqEnd;
	WORD	uESBQReqHead;
	WORD	uESBQReqTail;
	WORD	uESBQConfirmStart;
	WORD	uESBQConfirmEnd;
	WORD	uESBQConfirm;
    MADDR   BSSID;
	BYTE	LLCType;
	BYTE	NetworkType;
    BYTE    ESSID[34];
    BYTE    KeepESSID[34];
	BYTE	EtherType;
    BYTE    Channel;
    BYTE    DefaultChannel;
    WORD    StartSegment;
    WORD    BSSCount;
    WORD    JoinStaBSS;
    BYTE    CapablilityInfo;
    BYTE    AdHocTimes;
    BOOL    CardStart;
    ScanConf BSSset[20];
    BYTE    DriverState;
    BYTE    FrequencyDomain;
    BYTE    Version[2];
#ifdef LINUX_DRIVER
	struct	enet_statistics stat;	/* linux/if_ether.h */
#	ifdef WL_PCMCIA
	dev_node_t node;			/* PCMCIA only */
#	endif /* End of WL_PCMCIA */
#endif /* End of LINUX_DRIVER */

	WHEN_WLAPI_DEBUG(DWORD	ulDbgStat);
} WL_Adaptor, *PWL_Adaptor;

/*
 * API for accessing WL24xx SUTRO
 */
int  TraceNothing(const char *cpszFormat, ...);
void TraceBuf(void *pBuf, int nSize);
BOOL WL_GetFlashMacAddress(WL_Adaptor *pAdaptor);
void WL_FlashOutB(WL_Adaptor *pAdaptor, WORD uPage, WORD uAddr, BYTE cData);
void WL_FlashOutB128(WL_Adaptor *pAdaptor, WORD uPage, WORD uAddr, char *pbuf);
BYTE WL_FlashInB(WL_Adaptor *pAdaptor, WORD uPage, WORD uAddr);
void WL_FlashInSB(WL_Adaptor *pAdaptor, WORD uPage, WORD uSrc, void *pDest, int nSize);
WORD WL_GetFlashID(WL_Adaptor *pAdaptor);
BOOL WL_FlashWriteOk(WL_Adaptor *pAdaptor);
BOOL WL_FlashEraseSector(WL_Adaptor *pAdaptor, WORD uSector);
BOOL WL_FlashWriteByte(WL_Adaptor *pAdaptor, WORD uPage, WORD uAddr, BYTE cData);
BOOL WL_FlashWriteByte128(WL_Adaptor *pAdaptor, WORD uPage, WORD uAddr, char *pbuf);
BOOL WL_WriteFlash(WL_Adaptor *pAdaptor,unsigned char *pbuf,WORD uLen);
BOOL WL_BlockInterrupt(WL_Adaptor *pAdaptor);
BOOL WL_UnBlockInterrupt(WL_Adaptor *pAdaptor);
BOOL WL_HoldSutro(WL_Adaptor *pAdaptor);
BOOL WL_UnHoldSutro(WL_Adaptor *pAdaptor);
void WL_InteruptWla(WL_Adaptor *pAdaptor);
void WL_SetToWla(WL_Adaptor *pAdaptor, WORD uDest, void *pSrc, int nSize);
void WL_GetFromWla(WL_Adaptor *pAdaptor, WORD uSrc, void *pDest, int nSize);
void WL_SetDescNext(WL_Adaptor *pAdaptor, WORD uDesc, DWORD ulNext);
DWORD WL_GetDescNext(WL_Adaptor *pAdaptor, WORD uDesc);
WORD WL_GetTxDesc(WL_Adaptor *pAdaptor);
void WL_FreeTxDesc(WL_Adaptor *pAdaptor, WORD uDesc);
BOOL WL_LockSutro(WL_Adaptor *pAdaptor);
void WL_UnlockSutro(WL_Adaptor *pAdaptor);
WORD WL_GetTxBuffer(WL_Adaptor *pAdaptor, WORD uDataLen);
void WL_FreeTxBuffer(WL_Adaptor *pAdaptor, WORD uPtr);
BOOL WL_EnableReceiver(WL_Adaptor *pAdaptor);
BOOL WL_InitTx(WL_Adaptor *pAdaptor);
void WL_GetMibValue(WL_Adaptor *pAdaptor, WORD cIndex);
void WL_SetMibValue(WL_Adaptor *pAdaptor, WORD cIndex, void *pSrc, BYTE cSize);
BOOL WL_ResetBoard(WL_Adaptor *pAdaptor);
BOOL WL_InitFirmware(WL_Adaptor *pAdaptor);
BOOL WL_SendPkt(WL_Adaptor *pAdaptor, BYTE *pData, WORD uLen);
WORD WL_Receive(WL_Adaptor *pAdaptor, BYTE *pBuf, WORD uSize);
BOOL WL_ReceiveDone(WL_Adaptor *pAdaptor);
int  WL_ReceiveDataLook(WL_Adaptor *pAdaptor);
WORD WL_ReceiveData(WL_Adaptor *pAdaptor, BYTE *pBuf, WORD uSize);
BOOL WL_ReceiveDataDone(WL_Adaptor *pAdaptor);
void WL_Roaming(WL_Adaptor *pAdaptor);
BOOL WL_ESBQRequestTest(WL_Adaptor *pAdaptor);
void WL_ESBQRequest(WL_Adaptor *pAdaptor, WORD *ptr);
void WL_ESBQRequestFree(WL_Adaptor *pAdaptor);
BOOL WL_ESBQConfirm(WL_Adaptor *pAdaptor);
void WL_ESBQConfirmDone(WL_Adaptor *pAdaptor);
BOOL MgmtStart(WL_Adaptor *pAdaptor);
BOOL MgmtScan(WL_Adaptor *pAdaptor,WORD ChannelTime);
BOOL MgmtJoin(WL_Adaptor *pAdaptor,WORD stas);
BOOL MgmtAuthetication(WL_Adaptor *pAdaptor);
BOOL MgmtAssociation(WL_Adaptor *pAdaptor);
BOOL MgmtResync(WL_Adaptor *pAdaptor);
void MgmtScanConfirm(WL_Adaptor *pAdaptor,WORD Address);
void MgmtJoinConfirm(WL_Adaptor *pAdaptor,WORD Address);
void MgmtResyncConfirm(WL_Adaptor *pAdaptor);

#endif /* __WLAPI_H__ */
#ifndef __WL_IOCTL_H__
#define __WL_IOCTL_H__


#ifdef __FreeBSD__
/*
#  include <sys/ioctl.h>
#  include <sys/socket.h>
#  include <net/if.h>
#  include <sys/sockio.h>
#  include <sys/proc.h>*/
#  define   SIOC_WL_IOCTL               _IOWR('i', 60, struct ifreq)
#else
#  include <linux/if.h>
#  include <linux/sockios.h>
#  define   SIOC_WL_IOCTL               SIOCDEVPRIVATE
#endif


/*
 * WL_IOCTL_BLK is put into ifreq.ifr_data which is a union (16 bytes)
 *
 * Szeof(WL_IOCTL_BLK) must be less than 16 bteys.
 *
 * Ref: linux/if.h (Linux) and net/if.h (FreeBSD)
 */

typedef struct _WL_IOCTL_BLK {
    WORD uCmd;                /* Command to run */
    WORD uLen;                /* Length of the data buffer */
    unsigned char  *pData;              /* Pointer to the data buffer */
} WL_IOCTL_BLK, *PWL_IOCTL_BLK;

typedef struct _WL_IOCTL_PARA {
    BYTE DefaultChannel;
    BYTE Channel;
    BYTE NetworkType;
    BYTE LLCType;
    BYTE ESSID[34];
    BYTE KeepESSID[34];
    BYTE Version[2];
    BYTE FrequencyDomain;
} WL_IOCTL_PARA, *PWL_IOCTL_PARA;

#define WL_IOCTL_GET_PARAMETER   0 /* Get parameter */
#define WL_IOCTL_SET_PARAMETER   1 /* Get parameter */
#define WL_IOCTL_WRITE_FLASH     2 /* Write firmware into Flash ROM */
#define WL_IOCTL_SET_RESET       3 /* Reset the firmware */

/*
 * For not including wlapi.h, because wlu.c does not need it!
 */
struct device;  /* defined only in __KERNEL__ mode */
struct _WL_Adaptor;

int WL_Ioctl(struct device *dev, struct ifreq *rq, int cmd);
int WL_IoctlWla(struct _WL_Adaptor *pAdaptor, WL_IOCTL_BLK *pBlk, void *pDevice);

#endif /* __WL_IOCTL_H__ */
#ifndef __WLPNP_H__
#define __WLPNP_H__

/*
 * ---------------------------------------------------------------------------
 * Plug'n'play routines (Should be separated into another file)
 * ---------------------------------------------------------------------------
 */


#define PNP_ADDR_PORT       0x279
#define PNP_WR_PORT         0xA79

#define WL_VENDOR_ID        0x20246D68L

#define DEV_NUM_MAX         64

void NopLoop(void);
BOOL CompareCheckSum(BYTE Data[]);
void GetResourceData(BYTE ResourceData[], int nPort,int nCSN);
void GetSerialID(BYTE SerialID[], int nPnPRdPort);
int  Find_ReadDataPort(void);
int  Find_TotalCSN(int nPort);
void WL_GetDeviceInfo(int CSN,int nPnPReadPort, DWORD *pulBaseAddr, BYTE *pcIRQ);
void SetupPnPCard(int CSN, int irq, int IOBASE);
BOOL WL_PnpProbe(int nCardIndex, DWORD *pulBaseAddr, BYTE *pcIRQ);
void WL_PnpSetNonInitCard(DWORD ulBaseAddr, BYTE cIRQ);

#endif /* __WLPNP_H__ */

#ifndef __WL_24_H__
#define __WL_24_H__

void WL_RxInterrupt(struct device *dev);
void WL_Interrupt(int irq, void *dev_id);
int  WL_Reset(struct device *dev);
int  WL_StartXmit(struct sk_buff *skb, struct device *dev);
struct enet_statistics *WL_GetStats(struct device *dev);
int  WL_Close(struct device *dev);
int  WL_Open(struct device *dev);
void WL_SetMulticastList(struct device *dev);
int  WL_Probe(struct device *dev);

extern const char g_cpszVersion[];
extern const char *WL_OP_MODE_TABLE[];
extern int  WL_RSSI;

#endif /* __WL_24_H__ */


#define ZWL_DEBUG 

/*
 * Should be determinated by "zwl.h" which is generated by 'config', 
 * but in order to distribute *.o only, I assign the Max value. Too Bad!!
 * Note: currently each zwl_softc is about 340 bytes.
 */
/* #include "zwl.h" */
#undef  NZWL
#define NZWL    4

/* Always need pccard */
/* #include "card.h" */
/* #undef  NCARD */
#define NCARD  1 

#define ZWL_NO_PROBE_ISA    0x0001
#define ZWL_DBG_IF          0x0010
#define ZWL_DBG_PC          0x0020
#define ZWL_DBG_PERF        0x0040
#define ZWL_DBG_ETH         0x0080

#ifdef  ZWL_DEBUG
#  define TRACE_IF(format, args...)  \
            { if (zwl_flags & ZWL_DBG_IF) printf(format , ## args); }
#  define TRACE_PC(format, args...)   \
            { if (zwl_flags & ZWL_DBG_PC) printf(format , ## args); }
#  define TRACE_PERF(format, args...)   \
            { if (zwl_flags & ZWL_DBG_PERF) printf(format , ## args); }
#  define TRACE_ETH(format, args...)   \
            { if (zwl_flags & ZWL_DBG_ETH) printf(format , ## args); }
#  define WHEN_ZWL_DEBUG(x) x
#else
#  define TRACE_IF(format, args...)
#  define TRACE_PC(format, args...)
#  define TRACE_PERF(format, args...)
#  define TRACE_ETH(format, args...)
#  define WHEN_ZWL_DEBUG(x)
#endif

/*
#include <i386/isa/wlapi.h>
#include <i386/isa/wlpnp.h>
*/
/*
 * Uncomment this line if you would like to provide wlu support.
 * Wlu (Unix version) use ioctl() to control WL24xx cards.
 */
/*
#include <i386/isa/wlioctl.h>
*/
#define ZWL_BUFLEN    0x0670

static int zwl_probe_isa(struct isa_device *);  /* for zwldriver */
static int zwl_attach(struct isa_device *);     /* for zwldriver */

/*
 * These are used at boot time by the configuration program.
 * Refer: /usr/src/sys/i386/isa/isa_device.h
 */
struct isa_driver zwldriver = {
    zwl_probe_isa,  /* test whether device is present */
    zwl_attach,     /* setup driver for a device */
    "twl",          /* device name */
    0               /* true if other probes confuse us */
};

/*
 * Ethernet software status per interface.
 */
static struct zwl_softc {
    struct arpcom sc_arpcom;    /* Ethernet common part */
    WL_Adaptor  adaptor;

    u_char  *zwl_outbuf;        /* Outgoing packet buffer */
    u_char  gone;               /* PCCARD had been removed */

    WHEN_ZWL_DEBUG(DWORD   ulInterrupts); /* For debugging only */
    WHEN_ZWL_DEBUG(DWORD   ulRxNoMBuffer);
    WHEN_ZWL_DEBUG(DWORD   ulTxBcopyCalled);
    WHEN_ZWL_DEBUG(DWORD   ulRecvFunCalled);
    WHEN_ZWL_DEBUG(DWORD   ulZwlStartCalled);
    WHEN_ZWL_DEBUG(DWORD   ulZwlStartSendOk);
} zwl_softc[NZWL];

static const char *OP_MODE_TABLE[] = {
    "Access Point",
    "Infrastructure",
    "Ad-hoc",
    "Inter AP",
    NULL
};

static int zwl_flags;

/*
 * Definition for printing version information
 */
#ifdef WLAPI_DEBUG
#  define   WLAPI_DEBUG_STR         "WLAPI_DEBUG "
#else
#  define   WLAPI_DEBUG_STR         ""
#endif

#ifdef WL_SLOWDOWN_IO
#  define   WL_SLOWDOWN_IO_STR      "WL_SLOWDOWN_IO "
#else
#  define   WL_SLOWDOWN_IO_STR      ""
#endif

#ifdef ZWL_DEBUG
#  define   ZWL_DEBUG_STR           "ZWL_DEBUG "
#else
#  define   ZWL_DEBUG_STR           ""
#endif

#if NCARD > 0
#  define   NCARD_STR               "PCMCIA "
#else
#  define   NCARD_STR               ""
#endif

const char g_cpszVersion[] = "if_zwl.c: v1.25 " \
    WLAPI_DEBUG_STR WL_SLOWDOWN_IO_STR ZWL_DEBUG_STR NCARD_STR __DATE__;

int zwl_reset(struct zwl_softc *sc);            /* for wlioctl.c */

/*
 * BSD only provide bzero in kernel. :-(
 */
void* memset(void *pBuf, int nValue, size_t nSize)
{
    int i;
    for (i = 0; i < nSize; i++)
        ((BYTE *) pBuf)[i] = nValue;
        
    return pBuf;
}

/*
 * Create a null-terminated string containing the ESSID
 */
char *twl_ESSID(char *essid)
  {
  static char ess[34];
  int i;
  memset(ess, 0, 34);

  for (i = 0; i < (int)essid[1]; i++)
    {
    ess[i] = essid[i + 2];
    }
  ess[i] = 0;
  return ess;
  }

/*
 * Convert mbufs to buf, avoid additional copy if possible.
 *
 * WARNING: the return buffer may not start with ptr
 */
static caddr_t zwl_mtobuf(struct mbuf *m0, caddr_t pExtraBuf) {
    struct  mbuf *m;
    caddr_t pBuf;

    /* A trick to prevent bcopy */
    if (m0->m_next == NULL) {
        /* only one m, send it directly */
        pBuf = mtod(m0, caddr_t);
    }
    else {
        caddr_t ptr = pExtraBuf;
        
        /* a mbuf link-list, copy to buffer first */
        for (m = m0; m != 0; m = m->m_next) {
            bcopy(mtod(m, caddr_t), ptr, m->m_len);
            ptr += m->m_len;
        }
            
        pBuf = pExtraBuf;
    }
    
    return pBuf;
}

/*
 * Start to remove mbuf from queue and deliver it to the SUTRO
 * Note: Because we cannot expect how many bytes available in card,
 *       we don't actually remove an mbuf from queue until the packet
 *       had been sent to card successfully.
 */

#define Splimp() splimp()
#define Splx(s)  splx(s)

static void zwl_start(struct ifnet *ifp)
{
    register struct zwl_softc *sc = ifp->if_softc;
    struct  mbuf *m0;
    int     pktlen, s;
    caddr_t pBuf;
    static a_done = 0;
    WL_Adaptor *pAdaptor = &sc->adaptor;
    BOOL    bEnabled, bSend;
    if(! pAdaptor->CardStart) {
	ifp->if_timer = a_done = 0; /* reset watchdog */
	return;
	}

    /* 
       if a_done = 1, and if_timer = 0, then the watchdog has already been
       trigered, and is safe to set a_done to zero, so the watchdog can be
       set again if needed. 
    */
    if (ifp->if_timer == 0) a_done = 0; /* reset watchdog */

    WHEN_ZWL_DEBUG(sc->ulZwlStartCalled++);
    if (sc->gone)   /* For PCCARD: it had been removed */
        return;
    
    s = Splimp();
    
    /* Don't transmit if interface is busy or not running */
/*    if ((ifp->if_flags & (IFF_RUNNING|IFF_OACTIVE)) != IFF_RUNNING) {
        Splx(s);
        return;
    }*/

    if (ifp->if_flags & IFF_OACTIVE) {
	Splx(s);
	return;
    }

    /* Set IFF_OACTIVE immediately after checking */
    ifp->if_flags |= IFF_OACTIVE;

    /* Sneak a peek at the next packet */
    m0 = ifp->if_snd.ifq_head;
    if (m0 == 0) {
        ifp->if_flags &= ~IFF_OACTIVE;
        Splx(s);
        return;
    }

    /* We need to use m->m_pkthdr.len, so require the header */
    if ((m0->m_flags & M_PKTHDR) == 0)
        panic("zwl_start: no header mbuf");

    /* Try to padding the packet even if it's too short */
    pktlen = max(m0->m_pkthdr.len, ETHER_MIN_LEN);

    if (pktlen <= (int) sizeof(struct ether_header) || pktlen > ETHER_MAX_LEN) {
        TRACE_ETH("twl#: Tx <<<<< Invliad Ethernet packet (size=%d)\n", pktlen);
        /* packet is obviously too large: toss it */
        IF_DEQUEUE(&ifp->if_snd, m0);
         m_freem(m0);
        ifp->if_oerrors++;
        ifp->if_flags &= ~IFF_OACTIVE;
        Splx(s);
        return;
    }

    TRACE_IF("twl#: Tx .....\n");

    /* Mask interrupts from the SUTRO !! Do it as short as possible.        */
    /* We must mask interrupt from the same card, to prevent interrupt      */
    /* routine from accessing data structure and I/O port during WL_SendPkt */
    /* is running. It's very important!! */
    bEnabled = WL_BlockInterrupt(pAdaptor);

    pBuf = zwl_mtobuf(m0, sc->zwl_outbuf);

    /* Send the packet with default speed */
    bSend = WL_SendPkt(pAdaptor, pBuf, pktlen);
    
    /* Turn SUTRO interrupt back on only if it is originally enabled */
    /* Must check first, because zwlintr() calls zwl_start() too! */
    if (bEnabled)
        WL_UnBlockInterrupt(pAdaptor);
    
    /* If sent successfully turn off IFF_OACTIVE. Otherwise, buffer is all used */
    /* IFF_OACTIVE will be 0 again when the SUTRO interrupts us and return ISR_TX */
	{ /* <-- Make a point of getting rid of that, as it is useless. */  
        WHEN_ZWL_DEBUG(sc->ulZwlStartSendOk++);
/*	bSend=1;*/
        if (bSend)  /* Packet was sent, dequeue and disable watchdog. */
		{
		IF_DEQUEUE(&ifp->if_snd, m0);

#if NBPFILTER > 0
        	if (ifp->if_bpf)
            		bpf_mtap(ifp, m0);
#endif

        	TRACE_IF("twl#: Tx freem\n");
        	m_freem(m0);
        	ifp->if_timer = 0;  /* disable timer, packet was sent */
		a_done = 0; /* no timer in progress */
		/* printf("zwl_start: reset watchdog, send success.\n");*/
		}
	else 
	   {
	   if (!a_done) /* Packet was not sent seccessfuly */
		{
		/*printf("zwl_start: start watchdog, send failed,\n");*/
		ifp->if_timer = 5;  /* watchdog timeout 5 seconds */
		a_done = 1;  /* timer is in progress */
		}
	   /*else printf("zwl_start: send failed, watchdog already set.\n");*/
	   }
        ifp->if_flags &= ~IFF_OACTIVE;  /* Let others could own the transmitter */
    }
    /* I don't loop again. With field test, loop makes longer latency. */
    Splx(s);
}

static void zwl_old_start(struct ifnet *ifp)
{
    register struct zwl_softc *sc = ifp->if_softc;
    struct  mbuf *m0;
    int     pktlen, s;
    caddr_t pBuf;
    static a_done = 0;
    WL_Adaptor *pAdaptor = &sc->adaptor;
    BOOL    bEnabled, bSend;
    if(! pAdaptor->CardStart)
	{
	ifp->if_timer = a_done = 0; /* reset watchdog */
	return;
	}

    /*
       if a_done = 1, and if_timer = 0, then the watchdog has already been
       trigered, and is safe to set a_done to zero, so the watchdog can be
       set again if needed.
    */
    if (ifp->if_timer == 0) a_done = 0; /* reset watchdog */

    WHEN_ZWL_DEBUG(sc->ulZwlStartCalled++);
    if (sc->gone)   /* For PCCARD: it had been removed */
        return;

    s = Splimp();

    /* Don't transmit if interface is busy or not running */
/*    if ((ifp->if_flags & (IFF_RUNNING|IFF_OACTIVE)) != IFF_RUNNING) {
        Splx(s);
        return;
    }*/

    if (ifp->if_flags & IFF_OACTIVE) {
	Splx(s);
	return;
    }

    /* Set IFF_OACTIVE immediately after checking */
	ifp->if_flags |= IFF_OACTIVE;

    /* Sneak a peek at the next packet */
    m0 = ifp->if_snd.ifq_head;
    if (m0 == 0) {
        ifp->if_flags &= ~IFF_OACTIVE;
        Splx(s);
        return;
    }

    /* We need to use m->m_pkthdr.len, so require the header */
    if ((m0->m_flags & M_PKTHDR) == 0)
        panic("zwl_start: no header mbuf");

    /* Try to padding the packet even if it's too short */
    pktlen = max(m0->m_pkthdr.len, ETHER_MIN_LEN);

    if (pktlen <= (int) sizeof(struct ether_header) || pktlen > ETHER_MAX_LEN) {
        TRACE_ETH("twl#: Tx <<<<< Invliad Ethernet packet (size=%d)\n", pktlen);
        /* packet is obviously too large: toss it */
        IF_DEQUEUE(&ifp->if_snd, m0);
        m_freem(m0);
        ifp->if_oerrors++;
        ifp->if_flags &= ~IFF_OACTIVE;
        Splx(s);
        return;
    }

    TRACE_IF("twl#: Tx .....\n");

    /* Mask interrupts from the SUTRO !! Do it as short as possible.        */
    /* We must mask interrupt from the same card, to prevent interrupt      */
    /* routine from accessing data structure and I/O port during WL_SendPkt */
    /* is running. It's very important!! */
    bEnabled = WL_BlockInterrupt(pAdaptor);

    pBuf = zwl_mtobuf(m0, sc->zwl_outbuf);

    /* Send the packet with default speed */
    bSend = WL_SendPkt(pAdaptor, pBuf, pktlen);

    /* Turn SUTRO interrupt back on only if it is originally enabled */
    /* Must check first, because zwlintr() calls zwl_start() too! */
    if (bEnabled)
        WL_UnBlockInterrupt(pAdaptor);

    /* If sent successfully turn off IFF_OACTIVE. Otherwise, buffer is all used */
    /* IFF_OACTIVE will be 0 again when the SUTRO interrupts us and return ISR_TX */
	{ /* <-- Make a point of getting rid of that, as it is useless. */
        WHEN_ZWL_DEBUG(sc->ulZwlStartSendOk++);
/*	bSend=1;*/
        if (bSend)  /* Packet was sent, dequeue and disable watchdog. */
		{
		IF_DEQUEUE(&ifp->if_snd, m0);

#if NBPFILTER > 0
        	if (ifp->if_bpf)
            		bpf_mtap(ifp, m0);
#endif

        	TRACE_IF("twl#: Tx freem\n");
        	m_freem(m0);
        	ifp->if_timer = 0;  /* disable timer, packet was sent */
		a_done = 0; /* no timer in progress */
		/* printf("zwl_start: reset watchdog, send success.\n");*/
		}
	else
	   {
	   if (!a_done) /* Packet was not sent seccessfuly */
		{
		/*printf("zwl_start: start watchdog, send failed,\n");*/
		ifp->if_timer = 5;  /* watchdog timeout 5 seconds */
		a_done = 1;  /* timer is in progress */
		}
	   /*else printf("zwl_start: send failed, watchdog already set.\n");*/
	   }
        ifp->if_flags &= ~IFF_OACTIVE;  /* Let others could own the transmitter */
    }
    /* I don't loop again. With field test, loop makes longer latency. */
    Splx(s);
}

/*
 * convert buf into mbufs
 */
/* static */
static __inline struct mbuf *zwl_get(struct zwl_softc *sc, int pktlen)
{
    struct ifnet *ifp = &sc->sc_arpcom.ac_if;
    struct mbuf *m;
/*    struct mbuf *top, **mprev, *m;
      int len, offset; */

    ASSERT(pktlen >= 0);
    
    MGETHDR(m, M_DONTWAIT, MT_DATA);
    if (m == NULL)
        return NULL;

    m->m_pkthdr.rcvif = ifp;
    m->m_pkthdr.len = m->m_len = pktlen;
    if( pktlen > MHLEN) {
         MCLGET(m, M_DONTWAIT);
         if((m->m_flags & M_EXT) == 0) {
               m_freem(m);
               return NULL;
         }
    }

    WL_Receive(&sc->adaptor, mtod(m, caddr_t), pktlen);
    return m;

/*
    m->m_pkthdr.len = pktlen;
    len = MHLEN;
    top = NULL;
    mprev = &top;
    offset = 0;

    while (pktlen > 0) {
        if (top) {
            MGET(m, M_DONTWAIT, MT_DATA);
            if (m == NULL) {
                TRACE_PERF("twl#: No Rx mbuf available!\n");
                m_freem(top);
                return NULL;
            }
            len = MLEN;
        }
        if (pktlen >= MINCLSIZE) {
            MCLGET(m, M_DONTWAIT);
            if (m->m_flags & M_EXT)
                len = MCLBYTES;
        }
        m->m_len = len = min(pktlen, len);
        WL_Receive(&sc->adaptor, mtod(m, caddr_t), len);
        WHEN_ZWL_DEBUG(sc->ulRecvFunCalled++);
        offset += len;
        pktlen -= len;
        *mprev = m;
        mprev = &m->m_next;
    }
    ASSERT(top != NULL);
    return top;
*/
}

/*
 * Pass a packet up to the higher levels.
 */
static __inline int zwl_read(struct zwl_softc *sc, int pktlen, char *DAddress)
{
    struct ifnet *ifp = &sc->sc_arpcom.ac_if;
    WL_Adaptor *pAdaptor = &sc->adaptor;
    MADDR   *pAddr;
    
    struct mbuf *m;
    struct ether_header *eh;
    if (pktlen <= (int) sizeof(struct ether_header) || pktlen > ETHER_MAX_LEN) {
        TRACE_ETH("twl#: Rx <<<<< Invliad Ethernet packet (size=%d)\n", pktlen);
        ifp->if_ierrors++;
        
        /* skip this packet */
        return TRUE;
    }
    
    /* Pull packet off interface. */
    m = zwl_get(sc, pktlen);

    if (m == NULL) {
        WHEN_ZWL_DEBUG(sc->ulRxNoMBuffer++);
        return TRUE;
    }
    memcpy((char *) m->m_data, DAddress, 12);

    ifp->if_ipackets++;

    /* We assume the header fit entirely in one mbuf. */
    eh = mtod(m, struct ether_header *);

    /* Comes from monitored site ? */
    
#if NBPFILTER > 0
    /*
     * Check if there's a BPF listener on this interface.
     * If so, hand off the raw packet to BPF.
     */
    if (ifp->if_bpf) {
        bpf_mtap(ifp, m);

        /*
         * Note that the interface cannot be in promiscuous mode if
         * there are no BPF listeners.  And if we are in promiscuous
         * mode, we have to check if this packet is really ours.
         */
        if ((ifp->if_flags & IFF_PROMISC) &&
            (eh->ether_dhost[0] & 1) == 0 && /* !mcast and !bcast */
            bcmp(eh->ether_dhost, sc->sc_arpcom.ac_enaddr,
                ETHER_ADDR_LEN) != 0) {
            m_freem(m);
            return TRUE;
        }
    }
#endif

    /* We assume the header fit entirely in one mbuf. */
    m_adj(m, sizeof(struct ether_header));
    ether_input(ifp, eh, m);
    
    return TRUE;
}

static void zwl_rxintr(struct zwl_softc *sc)
{
    int         nSize,i;
    BOOL        bMorePkts;
    WL_Adaptor *pAdaptor = &sc->adaptor;
    struct ifnet *ifp = &sc->sc_arpcom.ac_if;
    MADDR   *pAddr;
	WORD	Address,cTemp,cTemp1,cTemp2,pkt_len;
    BYTE    Buff[32];
    BYTE    SignalID;
	MdCon	pMdConf;
	MdInd	pMdInd;
	GetConf pGetConf;
	StartConf pStartConf;
    AutheConf pAutheConf;
    AssocConf pAssocConf;
    
    struct mbuf *m;
    struct ether_header *eh;
    
    if (sc->gone)   /* For PCCARD: it had been removed */
        return;
        
    /* Check for data queue */
    bMorePkts = TRUE;
    while (bMorePkts) {
        TRACE_IF("twl#: Rx >>>>> Data\n");
        /* Pass a packet up to the higher levels */
		bMorePkts = FALSE;
		if (WL_ESBQConfirm(pAdaptor)){
			WL_GetFromWla(pAdaptor, pAdaptor->uESBQConfirm, &Address, sizeof(Address));
            WL_GetFromWla(pAdaptor, Address+2, &SignalID, sizeof(SignalID));
            switch (SignalID){
                case DeautheIndicate:
                    if (pAdaptor->NetworkType == Infrastructure){
                        printf("twl%d: Link has been lost. (DeautheIndicate)\n", ifp->if_unit);
	        	ifp->if_timer = 0;  /* disable timer */
                        pAdaptor->CardStart = FALSE;
                        for (i=0;i<3000;i++){
                            if (MgmtResync(pAdaptor)) break;
                        }
                    }
                    break;
                case DisassocIndicate:
                    if (pAdaptor->NetworkType == Infrastructure){
                        printf("twl%d: Link has been lost. (DosassocIndicate)\n", ifp->if_unit);
	        	ifp->if_timer = 0;  /* disable timer */
                        pAdaptor->CardStart = FALSE;
                        for (i=0;i<3000;i++){
                            if (MgmtResync(pAdaptor)) break;
                        }
                    }
                    break;
                case Alarm:
                    if (pAdaptor->NetworkType == Infrastructure){
                        printf("twl%d: Link has been lost. (Alarm)\n", ifp->if_unit);
	        	ifp->if_timer = 0;  /* disable timer */
                        pAdaptor->CardStart = FALSE;
                        for (i=0;i<3000;i++){
                            if (MgmtResync(pAdaptor)) break;
                        }
                    }
                    break;
		case MdConfirm:
			WL_GetFromWla(pAdaptor, Address, &pMdConf, sizeof(MdCon));
			WL_FreeTxBuffer(pAdaptor, pMdConf.Data);
			/*pAdaptor->CardStart = TRUE;*/
			zwl_start(ifp);
			break;

		case MdIndicate:
			WL_GetFromWla(pAdaptor, Address, &pMdInd, sizeof(MdInd));
			pAdaptor->StartSegment=pMdInd.Data;
			if (pAdaptor->LLCType == 1){
			WL_GetFromWla(pAdaptor, pMdInd.Data+sizeof(RxHeader)-6, &cTemp, sizeof(cTemp));
			WL_GetFromWla(pAdaptor, pMdInd.Data+sizeof(RxHeader)-4, &cTemp1, sizeof(cTemp1));
			WL_GetFromWla(pAdaptor, pMdInd.Data+sizeof(RxHeader)-2, &cTemp2, sizeof(cTemp2));
			if ((cTemp == 0xaaaa) && (cTemp1 == 0x03) && (cTemp2 == 0x0)){
				pkt_len = pMdInd.Size+12-24-4-6;
				pAdaptor->EtherType=EtherII;
			}
			else if (cTemp == 0xe0e0){
				pkt_len = pMdInd.Size+12-24-4+2;
				pAdaptor->EtherType=Ether802_3e;
			}
			else{
				pkt_len = pMdInd.Size+12-24-4+2;
				pAdaptor->EtherType=Ether802_3f;
			}
		}
		else{
			pkt_len = pMdInd.Size-24-4;
		}
	nSize=pkt_len;
/* Old code start */
	zwl_read(sc, nSize, (char *)&(pMdInd.DAddress));

/* Old code end */
				break;
				case GetConfirm:
					WL_GetFromWla(pAdaptor, Address, &pGetConf, sizeof(GetConf));
					if	(pGetConf.MibStatus == 0){
						switch (pGetConf.MibAtrib){
							case aMacAddress:
								break;
							case aCurrentRegDomain:
								break;
							default: break;
						}
					}
					break;
				case StartConfirm:
					WL_GetFromWla(pAdaptor, Address, &pStartConf, sizeof(StartConf));
                    if (pStartConf.Status == SUCCESS) pAdaptor->CardStart = TRUE;
                    else pAdaptor->CardStart = FALSE;
                    break;
                case ScanConfirm:
                    MgmtScanConfirm(pAdaptor,Address);
                    break;
                case JoinConfirm:
                    MgmtJoinConfirm(pAdaptor,Address);
                    break;
                case AssocConfirm:
                    WL_GetFromWla(pAdaptor, Address, &pAssocConf, sizeof(AssocConf));
                    if (pAssocConf.Status == SUCCESS){
                        printf("twl%d: Join BSS: ", ifp->if_unit); 
			printf("%02x:%02x:%02x:%02x:%02x:%02x, ESSID: %s, Channel: %d\n",
				pAdaptor->BSSID.b0,pAdaptor->BSSID.b1,
				pAdaptor->BSSID.b2,pAdaptor->BSSID.b3,
				pAdaptor->BSSID.b4,pAdaptor->BSSID.b5,
				twl_ESSID(pAdaptor->KeepESSID), pAdaptor->Channel);
                        pAdaptor->CardStart = TRUE;
                    }
                    else{
                        pAdaptor->CardStart = FALSE;
                    }
                    break;
                case AutheConfirm:
                    WL_GetFromWla(pAdaptor, Address, &pAutheConf, sizeof(AutheConf));
                    if (pAutheConf.Status == SUCCESS){
                        for (i=0;i<3000;i++){
                            if (MgmtAssociation(pAdaptor)) break;
                        }
                    }
                    else{
                        for (i=0;i<3000;i++){
                            if (MgmtResync(pAdaptor)) break;
                        }
                    }
                    break;
                case ResyncConfirm:
                    MgmtResyncConfirm(pAdaptor);
                    break;
				default: break;
			}
			WL_ESBQConfirmDone(pAdaptor);
			bMorePkts = TRUE;
		}
		/* free request if necessary */
		WL_ESBQRequestFree(pAdaptor);
    }
}

void zwlintr(int unit)
{
    struct zwl_softc *sc = &zwl_softc[unit];
    struct ifnet *ifp = &sc->sc_arpcom.ac_if;
    
    WL_Adaptor *pAdaptor = &sc->adaptor;
    BYTE        cTmp;
    BYTE        cStatus;
    int         x;

    if (sc->gone)   /* For PCCARD: it had been removed */
        return;

    /*printf("if_flags = %d (0x%x)\n", ifp->if_flags, ifp->if_flags);*/
    
    
    x = splhigh();

    /* Ack Interrupt */
    OutB(GCR_ECINT, pAdaptor->nBaseAddr + NIC_GCR);

    WL_BlockInterrupt(pAdaptor);
    
    /* For debugging */
    TRACE_IF("WL: interrupt %lu begin...\n", ++sc->ulInterrupts);
 
    zwl_rxintr(sc);

    /* Clean Zone End */
    WL_UnBlockInterrupt(pAdaptor);

    zwl_start(ifp);

    splx(x);
    return;
}

static void zwl_init(struct zwl_softc *sc)
{
    struct ifnet *ifp = &sc->sc_arpcom.ac_if;
    WL_Adaptor  *pAdaptor = &sc->adaptor;
    BYTE        cFilter, cCCR;
    int         s,i;

    if (sc->gone)   /* For PCCARD: it had been removed */
        return;

    /* For PCCARD */
    s = splhigh();

    /* Initial WL24 firmware */
    /* printf("twl%d: Initialize WL24xx firmware...\n", ifp->if_unit); */
    if (!WL_InitFirmware(pAdaptor)) {
        printf("twl%d: Cannot Initial Firmware!!\n", ifp->if_unit);
        /* Free IRQ, and mark IRQ as unused */
	splx(s);
        return;
    }

    if (sc->zwl_outbuf == NULL)
        sc->zwl_outbuf = malloc(ZWL_BUFLEN, M_TEMP, M_NOWAIT);

    /* Interface is now `running', with no output active. */
    ifp->if_flags |= IFF_RUNNING;
    ifp->if_flags &= ~IFF_OACTIVE;

    pAdaptor->CardStart=FALSE;
    pAdaptor->AdHocTimes=0; 
    /* Acknowledge Interrupt, for cleaning last state */
    OutB(GCR_ECINT, pAdaptor->nBaseAddr + NIC_GCR);
    
    /* Enable interrupt from card after all */
    WL_UnBlockInterrupt(pAdaptor);
    
    for (i=0;i<3000;i++){
	if (MgmtScan(pAdaptor,100)) break;
    } 
    printf("twl%d: device opened\n", ifp->if_unit);

    /* Attempt to start output, if any. */
    zwl_start(ifp);
    
    splx(s);
}

static void zwl_stop(struct zwl_softc *sc)
{
    struct ifnet *ifp = &sc->sc_arpcom.ac_if;
    WL_Adaptor  *pAdaptor = &sc->adaptor;

    if (sc->gone)   /* For PCCARD: it had been removed */
        return;
        
    /* Ack Interrupt */
    OutB(GCR_ECINT, pAdaptor->nBaseAddr + NIC_GCR);

    /* Mask interrupts from the SUTRO */
    WL_BlockInterrupt(pAdaptor);
    
    /* Free IRQ, and mark IRQ as unused */
    
   printf("twl%d: device closed\n", ifp->if_unit); 
}

/*
 * Reset the SUTRO. Just simply stop and initial again
 */ 
int zwl_reset(struct zwl_softc *sc)
{
    int s;
    
    if (sc->gone)   /* For PCCARD: it had been removed */
        return -EIO;
        
    s = splimp();
    
    zwl_stop(sc);
    zwl_init(sc);
    
    splx(s);
    return 0;
}

static void zwl_watchdog(struct ifnet *ifp)
{
    struct zwl_softc *sc = ifp->if_softc;

    if (sc->gone)   /* For PCCARD: it had been removed */
        return;
        
    /* WebWorld: Uncommented the below three lines */
    log(LOG_ERR, "twl%d: watchdog timeout.\n", ifp->if_unit);
    sc->sc_arpcom.ac_if.if_oerrors++;
    zwl_reset(sc);
}

static void zwl_setfilter(struct zwl_softc *sc)
{
    struct ifnet *ifp = &sc->sc_arpcom.ac_if;
    WL_Adaptor  *pAdaptor = &sc->adaptor;
    BYTE cFilter;
    int s;

    if (ifp->if_flags & IFF_PROMISC) {
        /* Promiscuous mode */
/*        cFilter = RMR_UNICAST | RMR_PROMISCUOUS | RMR_ALL_MULTICAST | RMR_BROADCAST;*/
    }
    else if (sc->sc_arpcom.ac_multicnt > 0) {
        /* Allow multicast */
/*        cFilter = RMR_UNICAST | RMR_ALL_MULTICAST | RMR_BROADCAST; */
    }
    else {
        /* Normal mode */
/*        cFilter = RMR_UNICAST | RMR_BROADCAST; */
    }
    
    /* Mask all SW/HW interrupts) */
    s = splhigh();

    /* Must not be interrupted */
/*    WL_SetMibValue(pAdaptor, TYPE_EXTRA_MIB, IDX_RECEIVEMODE, &cFilter, sizeof(cFilter)); */
    
    splx(s);
}

static int zwl_ioctl(register struct ifnet *ifp, int command, caddr_t data)
{
    struct zwl_softc *sc = ifp->if_softc;
    register struct ifaddr *ifa = (struct ifaddr *) data;
    struct ifreq *ifr = (struct ifreq *) data;
    WL_Adaptor  *pAdaptor = &sc->adaptor;
    int s, error = 0;

    TRACE_PERF("twl#: ulInterrupts=%10lu,\tulRxNoMBuffer=%lu\n"
               "twl#: ulTxBcopyCalled=%10lu,\tulRecvFunCalled=%lu\n"
               "twl#: ulZwlStartCalled=%10lu,\tulZwlStartSendOk=%lu",
                sc->ulInterrupts, sc->ulRxNoMBuffer, 
                sc->ulTxBcopyCalled, sc->ulRecvFunCalled,
                sc->ulZwlStartCalled, sc->ulZwlStartSendOk);

    if (sc->gone) { /* For PCCARD: it had been removed */
        ifp->if_flags &= ~IFF_RUNNING;
        return ENXIO;
    }

    s = splnet();

    switch (command) {

      case SIOCSIFADDR:
      case SIOCGIFADDR:
      case SIOCSIFMTU:
	error = ether_ioctl(ifp, command, data);
	break;

    case SIOCSIFFLAGS:
        if ((ifp->if_flags & IFF_UP) == 0 &&
            (ifp->if_flags & IFF_RUNNING) != 0) {
            /* If interface is marked down and it is running, then stop it. */
            zwl_stop(sc);
            ifp->if_flags &= ~IFF_RUNNING;
        } else if ((ifp->if_flags & IFF_UP) != 0 &&
               (ifp->if_flags & IFF_RUNNING) == 0) {
            /* If interface is marked up and it is stopped, then start it. */
            TRACE("twl#: call zwl_init in zwlioctl\n");
            zwl_init(sc);
        } else {
            /* Deal with flags changes: IFF_MULTICAST, IFF_PROMISC */
            zwl_setfilter(sc);
        }
        break;
	case SIOCADDMULTI:
	case SIOCDELMULTI:
	    error = 0;
	    break;
 #ifdef notdef
      case SIOCGHWADDR:
	bcopy((caddr_t) sc->sc_addr, (caddr_t) & ifr->ifr_data,
	      sizeof(sc->sc_addr));
	break;
#endif

#ifdef __WL_IOCTL_H__
    case SIOC_WL_IOCTL: {
        /* Must block all interrupts */
        int hi = splhigh();
        struct ifreq *ifr = (struct ifreq *)data;
        error = -WL_IoctlWla(pAdaptor, (WL_IOCTL_BLK *) &ifr->ifr_data, sc);
        splx(hi);
        break;
    }
#endif

    default:
        error = EINVAL;
        break;
    }

    splx(s);
    return error;
}

static int zwl_attach(struct isa_device *isa_dev)
{
    int unit = isa_dev->id_unit;
    struct zwl_softc *sc = &zwl_softc[unit];
    struct ifnet *ifp = &sc->sc_arpcom.ac_if;
    int attached = (ifp->if_softc != NULL);
    
    WL_Adaptor  *pAdaptor = &sc->adaptor;
    int i;

    /* print probe information */
/*    printf("twl%d: WL24xx at 0x%3.3x, MAC address in flash ROM:",  */
    printf("twl%d: WebWorld WL24xx at 0x%3.3x < Address:",
            unit, pAdaptor->nBaseAddr);
    for (i = 0; i < 6; i++) {
        printf("%c%02x", i ? ':' : ' ', ((BYTE *) &pAdaptor->MacAddress)[i]);
    }
    printf(" >\n");

    bcopy(&pAdaptor->MacAddress, sc->sc_arpcom.ac_enaddr, ETHER_ADDR_LEN);

    /* Initialize ifnet structure. */
    ifp->if_softc = sc;
    ifp->if_unit = unit;
    ifp->if_name = "twl";
/*    ifp->if_mtu = ETHERMTU;*/
    ifp->if_output = ether_output;
    ifp->if_start = zwl_start;
    ifp->if_ioctl = zwl_ioctl;
    ifp->if_watchdog = zwl_watchdog;
    ifp->if_init = (if_init_f_t*) zwl_init;
    ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX;
    ifp->if_snd.ifq_maxlen = ifqmaxlen;
/*    ifp->if_baudrate = 10000000;*/
    
    /* Must have this, otherwise you'll get ENOBUFS in sendto(2) */
    /* It is specially for PCCARD, that doesn't initialize it.   */
    /* Refer: ifinit() in net/if.c (ifqmaxlen declared there)    */
    /* After 1 day's debugging... :-(                            */
/*  if (ifp->if_snd.ifq_maxlen == 0 )
        ifp->if_snd.ifq_maxlen = ifqmaxlen;*/
    
    /* Now we can attach the interface. (For PCCARD, must test first) */
    /* XXX: After 2 days debuging...                                  */
    if (!attached) {
        if_attach(ifp);
        ether_ifattach(ifp);
#if NBPFILTER > 0
        bpfattach(ifp, DLT_EN10MB, sizeof(struct ether_header));
#endif
    }
    /*printf("twl%d: twl_attach ok!\n", unit); */
    zwl_init(sc);
    TRACE("twl%d: zwl_attach ok!\n", unit);
    return 1; /* was 1 */
}

/*
 * Probe if there is any ISA WL24xx card.
 * Return: 0 if failed, or occupied I/O region
 *
 * To understand ISA probing procdure, refer /usr/src/sys/i386/isa/isa.c
 * CAUTION: id_irq is bitmask (refer /usr/src/sys/i386/isa/icu.h)
 */
static int zwl_probe_isa(struct isa_device * id)
{
    struct zwl_softc *sc = &zwl_softc[id->id_unit];
    int  nCardIndex = id->id_unit + 1;      /* start with 1 */
    
    WL_Adaptor  *pAdaptor = &sc->adaptor;
    int         i;

    /* Check if byte alignment is correct or not. FOR DEBUGGING ONLY */
    if (sizeof(TX_HEADER) != 54) {
        printf("zwl_probe_isa.c: compiler alignment option must be 1 byte.\n");
        return 0;
    }

    /* device zwl0 at isa? flags 0x1 net vector zwlintr */
    zwl_flags = id->id_flags;
    
    /* higher 16 bits as wlapi debug flags */
    WHEN_WLAPI_DEBUG(WL_DbFlag = zwl_flags >> 16);   
    
    if (id->id_flags & ZWL_NO_PROBE_ISA) {
        /* Specically for PCCARD */
        printf("twl%d: Skip WL24xx ISA Probe (flags=0x%x)\n", id->id_unit, id->id_flags);
        return 0;
    }
    
    /* port ?       => id_iobase = 0xffffffff = -1 */
    /* port 0xffff  => id_iobase = 0xffff */
    /* not assigned => id_iobase = 0 */
    if (id->id_iobase <= 0 || id->id_iobase == 0xffff || id->id_irq == 0) {
        DWORD   ulBaseAddr;
        BYTE    cIRQ;

        printf("twl%d: Probing for WebWorld PnP WL24xx adaptor.\n",id->id_unit);
        if (WL_PnpProbe(nCardIndex, &ulBaseAddr, &cIRQ)) {
            id->id_iobase   = ulBaseAddr;
            id->id_irq      = 1 << cIRQ;
            printf("twl%d: WebWorld WL24xx PnP card probe at 0x%3.3x, IRQ %d\n", 
                    id->id_unit, (int) id->id_iobase, (int) cIRQ);
        }
        else {
            printf("twl%d: Cannot probe PnP card, you must specify base address and irq.\n", id->id_unit);
            return 0;
        }
    }
    else {
        /* Try to find non-PNPed WL24xx PnP cards that are installed */
        /* in a PC which does not support PnP BIOS. */
        /* If found, set the card's IRQ and I/O up. e.g. ffs(0x0010)=5 */
        WL_PnpSetNonInitCard(id->id_iobase, ffs(id->id_irq)-1);
    }    
    
    sc->gone = 0;
    bzero(pAdaptor, sizeof(*pAdaptor));
    pAdaptor->nBaseAddr = id->id_iobase;

    if (!WL_GetFlashMacAddress(pAdaptor)) {
        printf("twl%d: Unable to get MAC address for adaptor.\n", id->id_unit);
        return 0;
    }
/* initila adaptor parameter // add by jss */
    pAdaptor->NetworkType = Infrastructure;
    pAdaptor->LLCType = 1;
    pAdaptor->DefaultChannel = 1;
    pAdaptor->BSSCount = 0;
    pAdaptor->JoinStaBSS = 0;
    pAdaptor->AdHocTimes = 0;
    pAdaptor->DriverState = 0;
    pAdaptor->CardStart = FALSE;
    pAdaptor->ESSID[0] = 0;
    pAdaptor->ESSID[1] = 3;
    pAdaptor->ESSID[2] = 'A';
    pAdaptor->ESSID[3] = 'N';
    pAdaptor->ESSID[4] = 'Y';
    switch (pAdaptor->FrequencyDomain){
        case 0x10:
            pAdaptor->DefaultChannel = 1;
            break;
        case 0x20:
            pAdaptor->DefaultChannel = 1;
            break;
        case 0x30:
            pAdaptor->DefaultChannel = 1;
            break;
        case 0x31:
            pAdaptor->DefaultChannel = 10;
            break;
        case 0x32:
            pAdaptor->DefaultChannel = 10;
            break;
        case 0x40:
            pAdaptor->DefaultChannel = 14;
            break;
        default:
            pAdaptor->DefaultChannel = 1;
            break;
    }
    printf("twl%d: ISA card probed\n", id->id_unit);
    return WL_IO_REGION;
}


#if NCARD > 0
/*
 *  PC-Card (PCMCIA) specific code.
 */

static int  zwl_init_pccard(struct pccard_devinfo *);   /* Initial device */
static void zwl_unload_pccard(struct pccard_devinfo *); /* Disable driver */
static int  zwl_intr_pccard(struct pccard_devinfo *);   /* Interrupt handler */
/*
static struct pccard_device zwl_info = {
    "twl",
    zwl_init_pccard,
    zwl_unload_pccard,
    zwl_intr_pccard,
    0,                  // Attributes - presently unused 
    &net_imask          // Interrupt mask for device 
                        // XXX - Should this also include net_imask? 
};*/

/* DATA_SET(pccarddrv_set, zwl_info);*/
PCCARD_MODULE(twl, zwl_init_pccard, zwl_unload_pccard, zwl_intr_pccard, 0, net_imask);

/*
 * Probe if there is any PCCARD WL24xx card.
 * Return: 0 if failed, or occupied I/O region
 *
 * Do what zwl_probe() does, but skip PnP functions
 */
static int zwl_probe_pccard(struct isa_device * id)
{
    struct zwl_softc *sc = &zwl_softc[id->id_unit];
    int  nCardIndex = id->id_unit + 1;      /* start with 1 */
    
    WL_Adaptor  *pAdaptor = &sc->adaptor;
    int         s;

    sc->gone = 0;
/*    bzero(sc, sizeof(struct zwl_softc));*/
    bzero(pAdaptor, sizeof(WL_Adaptor));
    pAdaptor->nBaseAddr = id->id_iobase;

    TRACE_PC("twl#: WL_GetFlashMacAddress()\n");
    s = splbio();
    if (!WL_GetFlashMacAddress(pAdaptor)) {
        printf("twl%d: Unable to get MAC address for adaptor.\n", id->id_unit);
        return 0;
    }

/* initila adaptor parameter // add by jss */
    pAdaptor->NetworkType = Infrastructure;
    pAdaptor->LLCType = 1;
    pAdaptor->DefaultChannel = 1;
    pAdaptor->BSSCount = 0;
    pAdaptor->JoinStaBSS = 0;
    pAdaptor->AdHocTimes = 0;
    pAdaptor->DriverState = 0;
    pAdaptor->CardStart = FALSE;
    pAdaptor->ESSID[0] = 0;
    pAdaptor->ESSID[1] = 3;
    pAdaptor->ESSID[2] = 'A';
    pAdaptor->ESSID[3] = 'N';
    pAdaptor->ESSID[4] = 'Y';
    switch (pAdaptor->FrequencyDomain){
        case 0x10:
            pAdaptor->DefaultChannel = 1;
            break;
        case 0x20:
            pAdaptor->DefaultChannel = 1;
            break;
        case 0x30:
            pAdaptor->DefaultChannel = 1;
            break;
        case 0x31:
            pAdaptor->DefaultChannel = 10;
            break;
        case 0x32:
            pAdaptor->DefaultChannel = 10;
            break;
        case 0x40:
            pAdaptor->DefaultChannel = 14;
            break;
        default:
            pAdaptor->DefaultChannel = 1;
            break;
    }
    splx(s);
    
    /*printf("twl%d: PCCARD card probed\n", id->id_unit);*/
    return WL_IO_REGION;
}

/*
 *  Initialize the device - called from Slot manager.
 *  Return ENXIO if success, or 0 if failed
 */
static int zwl_init_pccard(struct pccard_devinfo *devi)
{
    int unit = devi->isahd.id_unit;

    /* validate unit number. */
    if (unit >= NZWL) {
        printf("twl%d: Only assign %d units when compiling\n", unit, NZWL);
        return(ENODEV);
    }
    
    /* io base from isa_device, but irq from pccard_devinfo.slot? ref:slot.h */
    /*printf("twl%d: PCCARD base=0x%3.3x, irq=%d\n", 
            unit, devi->isahd.id_iobase, devi->slt->irq);*/
    
    /* WebWorld: changed == to = on below line */
    devi->isahd.id_irq = devi->slt->irq;
    if (zwl_probe_pccard(&devi->isahd) == 0)
        return(ENXIO);

    if (zwl_attach(&devi->isahd) == 0)
        return(ENXIO);

    return(0);
}

/*
 *  zwl_unload_pccard - unload the driver and clear the table.
 *  XXX TODO:
 *  This is usually called when the card is ejected, but
 *  can be caused by a modunload of a controller driver.
 *  The idea is to reset the driver's view of the device
 *  and ensure that any driver entry points such as
 *  read and write do not hang.
 */
static void zwl_unload_pccard(struct pccard_devinfo *devi)
{
    struct zwl_softc *sc = &zwl_softc[devi->isahd.id_unit];
    struct ifnet *ifp = &sc->sc_arpcom.ac_if;

    if (sc->gone) {
        printf("twl%d: already unloaded\n", devi->isahd.id_unit);
        return;
    }

    ifp->if_flags &= ~IFF_RUNNING;
    /* originally commented out */
    /*if_down(ifp); */
    sc->gone = 1;

    printf("twl%d: unload\n", devi->isahd.id_unit);
}

/*
 *  zwl_intr_pccard - Shared interrupt called from
 *  front end of PC-Card handler.
 */
static int zwl_intr_pccard(struct pccard_devinfo *devi)
{
    zwlintr(devi->isahd.id_unit);
    return(1);
}

#else   /* NCARD = 0 */
#   warning "************************"
#   warning "* WL24xx PCMCIA is off *"
#   warning "************************"
#endif  /* NCARD > 0 */
/*
 * WL24xx Wireless LAN Card Common APIs
 *
 * Design Issues:
 *
 *   1. Prevent using O/S dependent routines as possible.
 *   2. Fully utilize Rx/Tx queueing rings.
 *   3. Avoid additional memcpy between O/S and card if its size is large.
 *      (But if the overhead is little, I prefer to make the program clear.)
 *
 * References:
 *
 *   1. WL24xx packet drivers (tooasm.asm)
 *   2. Access Point Firmware Interface Specification for IEEE 802.11 SUTRO
 *   3. IEEE 802.11
 *
 * Additional Information:
 *
 *   1. To prevent forward declaration, I put primitive functions ahead.
 *   2. All veriables are named with type, with form <type><Name> e.g.:
 *      cTmpVar -> type is 'c', and Name is 'TmpVar'
 *
 *      Type List:
 *          c   - unsigned char (BYTE)
 *          n   - int           (DWORD)
 *          b   - boolean int   (BOOL, DWORD)
 *          u   - unsigned int  (WORD)
 *          ul  - unsigned long (DWORD)
 *          p   - pointer       (DWORD)
 *          psz - pointer to string ended with zero
 *          cpsz- const psz
 *          q   - queue (structure)
 *      Others are defined by Windows programming convensions.
 *
 * To do:
 *
 *   1. Fully compatiable with IEEE 802.11 (SUTRO need to be refined too)
 *   2. Process Broadcast Queue , PsPoll Queue, and CfPoll Queue
 *
 */


WHEN_WLAPI_DEBUG(DWORD WL_DbFlag = 0x00000000L);

static BYTE WL_FPAGE[] = { BSS_FPAGE0, BSS_FPAGE1, BSS_FPAGE2, BSS_FPAGE3 };

static const BYTE WL_SIGNAL_TABLE[2] = { 
    0x0A, 
    0x14   /* QPSK */
};


/*
 * For debugging
 */
int  TraceNothing(const char *cpszFormat, ...)
{
    return 0;
}

void TraceBuf(void *pBuf, int nSize)
{
    int     i;
    BYTE    *pCh = (BYTE *) pBuf;
    
    for (i = 0; i < nSize; i++)
        printk("%2.2x ", pCh[i]);
        
    printk("\n");
}

/*
 * Get Ethernet MAC addresss.
 *
 * WARNING: We switch to FPAGE0 and switc back again.
 *          Making sure there is no other WL function beening called by ISR.
 */
BOOL WL_GetFlashMacAddress(WL_Adaptor *pAdaptor)
{
    int   nBaseAddr = pAdaptor->nBaseAddr;
    
    ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_TRIVIAL_FUNC) == 0);
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_FLASH_READ);
    
    /* get MAC Address */
    OutB_P(BSS_FPAGE3, nBaseAddr + NIC_BSS);  /* BSS */
    OutB_P(0x00, nBaseAddr + NIC_LMAL);       /* LMAL */
    OutB_P(0x40, nBaseAddr + NIC_LMAH);       /* LMAH */

    /* wait for reading EEPROM */
    NOPLOOP(100);
    pAdaptor->MacAddress.b0 = InB(nBaseAddr + NIC_IODPA);
    NOPLOOP(100);
    pAdaptor->MacAddress.b1 = InB(nBaseAddr + NIC_IODPA);
    NOPLOOP(100);
    pAdaptor->MacAddress.b2 = InB(nBaseAddr + NIC_IODPA);
    NOPLOOP(100);
    pAdaptor->MacAddress.b3 = InB(nBaseAddr + NIC_IODPA);
    NOPLOOP(100);
    pAdaptor->MacAddress.b4 = InB(nBaseAddr + NIC_IODPA);
    NOPLOOP(100);
    pAdaptor->MacAddress.b5 = InB(nBaseAddr + NIC_IODPA);
    NOPLOOP(100);
    pAdaptor->FrequencyDomain = InB(nBaseAddr + NIC_IODPA);
    NOPLOOP(100);
    OutB_P(BSS_FPAGE0, nBaseAddr + NIC_BSS); 
    OutB_P(0x04, nBaseAddr + NIC_LMAL);
    OutB_P(0x40, nBaseAddr + NIC_LMAH);
    NOPLOOP(100);
    pAdaptor->Version[0] = InB(nBaseAddr + NIC_IODPA);
    NOPLOOP(100);
    pAdaptor->Version[1] = InB(nBaseAddr + NIC_IODPA);
    /* switch to SRAM Page 0 (for safety) */
    OutB(BSS_SPAGE0, pAdaptor->nBaseAddr + NIC_BSS);
    
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_FLASH_READ);

    /* Z-Com's MAC Address should be 00:60:... */
    return pAdaptor->MacAddress.b0 == 0x00 && pAdaptor->MacAddress.b1 == 0x60;
}

void WL_FlashOutB(WL_Adaptor *pAdaptor, WORD uPage, WORD uAddr, BYTE cData)
{
    /* switch to Flash RAM Page uPage */
    OutB(WL_FPAGE[uPage], pAdaptor->nBaseAddr + NIC_BSS);
    
    /* set LMAL and LMAH */
    OutB(uAddr & 0xff, pAdaptor->nBaseAddr + NIC_LMAL);
    OutB((uAddr >> 8) & 0x7f,   pAdaptor->nBaseAddr + NIC_LMAH);
    
    /* out data to Port A */
    OutB(cData, pAdaptor->nBaseAddr + NIC_IODPA);
}

void WL_FlashOutB128(WL_Adaptor *pAdaptor, WORD uPage, WORD uAddr, char *pbuf)
{
    int i;
    /* switch to Flash RAM Page uPage */
    OutB(WL_FPAGE[uPage], pAdaptor->nBaseAddr + NIC_BSS);
    
    /* set LMAL and LMAH */
    OutB(uAddr & 0xff, pAdaptor->nBaseAddr + NIC_LMAL);
    OutB((uAddr >> 8) & 0x7f, pAdaptor->nBaseAddr + NIC_LMAH);
    
    /* out data to Port A */
    for (i=0;i<128;i++){
       OutB(*pbuf, pAdaptor->nBaseAddr + NIC_IODPA);
       pbuf++;
    }
    NOPLOOP(10000);
}

BYTE WL_FlashInB(WL_Adaptor *pAdaptor, WORD uPage, WORD uAddr)
{
    /* switch to Flash RAM Page uPage */
    OutB(WL_FPAGE[uPage], pAdaptor->nBaseAddr + NIC_BSS);
    
    /* set LMAL and LMAH */
    OutB(uAddr & 0xff, pAdaptor->nBaseAddr + NIC_LMAL);
    OutB(uAddr >> 8,   pAdaptor->nBaseAddr + NIC_LMAH);
    
    /* out data to Port A */
    return InB(pAdaptor->nBaseAddr + NIC_IODPA);
}

void WL_FlashInSB(WL_Adaptor *pAdaptor, WORD uPage, WORD uSrc, void *pDest, int nSize)
{
    /* switch to SRAM Page 0 */
    OutB(WL_FPAGE[uPage], pAdaptor->nBaseAddr + NIC_BSS);
    
    /* set LMAL and LMAH */
    OutB(uSrc & 0xff, pAdaptor->nBaseAddr + NIC_LMAL);
    OutB(uSrc >> 8,   pAdaptor->nBaseAddr + NIC_LMAH);
    
    /* rep get from Port A */
    insb(pAdaptor->nBaseAddr + NIC_IODPA, pDest, nSize);
}

/*
 * When calling this function, must hold SUTRO first.
 */
WORD WL_GetFlashID(WL_Adaptor *pAdaptor)
{

    BYTE cByte0, cByte1;
    WORD uID;
    
    /* Autoselect command */
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xAA);
    WL_FlashOutB(pAdaptor, 0, 0x2AAA, 0x55);
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0x90);
    NOPLOOP(10000);
    
    cByte0 = WL_FlashInB(pAdaptor, 0, 0x0);
    cByte1 = WL_FlashInB(pAdaptor, 0, 0x1);
    
    uID = (cByte0 << 8) | cByte1;
    
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xAA);
    WL_FlashOutB(pAdaptor, 0, 0x2AAA, 0x55);
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0x0f0);
    NOPLOOP(10000);
/*    printk("Flash ROM ID = 0x%x\n", uID);  */
    return uID;
}

/*
 * Polling if Erase/Programming command is completed
 * Note: IF a==b THEN XOR(a,b) = 0 
 *
 * When calling this function, must hold SUTRO first.
 */
BOOL WL_FlashWriteOk(WL_Adaptor *pAdaptor)
{
    BYTE cByte0, cByte1;

    /* Check 'Toggle Bit' (DQ6) to see if completed */
    do {
        cByte0 = WL_FlashInB(pAdaptor, 0, 0x0);
        cByte1 = WL_FlashInB(pAdaptor, 0, 0x0);
        
        /* Test if exceeded Time Limits (DQ5) */
        if (cByte1 & 0x20) {
            /* Must test DQ6 again before return FALSE */
            cByte0 = WL_FlashInB(pAdaptor, 0, 0x0);
            cByte1 = WL_FlashInB(pAdaptor, 0, 0x0);
    
            return ((cByte0 ^ cByte1) & 0x40) == 0;
        }
            
    } while ((cByte0 ^ cByte1) & 0x40);
    
    return TRUE;
}

/*
 * When calling this function, must hold SUTRO first.
 */
BOOL WL_FlashEraseSector(WL_Adaptor *pAdaptor, WORD uSector)
{
    WORD uPage, uAddr;

    uPage = uSector / 2;
    uAddr = (uSector & 1) ? 0x4000 : 0x0;
    
    /* Sector Erase command (6 commands must within 100 uS) */
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xAA);
    WL_FlashOutB(pAdaptor, 0, 0x2AAA, 0x55);
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0x80);
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xAA);
    WL_FlashOutB(pAdaptor, 0, 0x2AAA, 0x55);
    WL_FlashOutB(pAdaptor, uPage, uAddr, 0x30);

    return WL_FlashWriteOk(pAdaptor);
}

/*
 * When calling this function, must hold SUTRO first.
 */
BOOL WL_FlashWriteByte(WL_Adaptor *pAdaptor, WORD uPage, WORD uAddr, BYTE cData)
{
    /* Autoselect command */
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xAA);
    WL_FlashOutB(pAdaptor, 0, 0x2AAA, 0x55);
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xA0);

    WL_FlashOutB(pAdaptor, uPage, uAddr, cData);

    return WL_FlashWriteOk(pAdaptor);
}

/*
 * When calling this function, must hold SUTRO first.
 */
BOOL WL_FlashWriteByte128(WL_Adaptor *pAdaptor, WORD uPage, WORD uAddr,char *pbuf)
{
    /* Autoselect command */
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xAA);
    WL_FlashOutB(pAdaptor, 0, 0x2AAA, 0x55);
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xA0);

    WL_FlashOutB128(pAdaptor, uPage, uAddr, pbuf);
    return TRUE;
}
/*
 * Write pAdaptor->mibExtra into Flash ROM
 *
 * In fact, only first 29 bytes are used. (not all of extra MIB)
 * To prevent alter other data will be used by future versions, we preserve
 * 256 bytes.
 */
BOOL WL_WriteFlash(WL_Adaptor *pAdaptor,unsigned char *pbuf, WORD uLen)
{
    int  i;
    WORD uFlashID;
    BYTE buf[256];
    BOOL bRunning;
    DWORD BufAddress;
    bRunning = WL_HoldSutro(pAdaptor);
    
    uFlashID = WL_GetFlashID(pAdaptor);
    
    BufAddress= *(DWORD *) pbuf;
    pbuf+=4;uLen-=4;
    if ((uFlashID == 0x0120) || (uFlashID == 0x016E)){
        /* It's AMD AM29F010, must be erase before programming */
        /* Erase 1st 16 Kbytes within Page 0 */
        if ((BufAddress & 0x3fff) ==0){
            if (!WL_FlashEraseSector(pAdaptor, BufAddress>>14 )) {
                printk("wlapi.c: WL_FlashEraseSector(0) failed!\n");
                if (bRunning)
                    WL_UnHoldSutro(pAdaptor);
                return FALSE;
            }
        }
        for (i = 0; i < uLen; i++) {
           if (!WL_FlashWriteByte(pAdaptor, BufAddress>>15, BufAddress+i, *pbuf)) {
               printk("wlapi.c: WL_FlashWriteByte(buf[%d]) failed!\n", i);
               if (bRunning)
                   WL_UnHoldSutro(pAdaptor);
               return FALSE;
           }
           pbuf++;
       }
       return TRUE;
    }
    else if (uFlashID == 0x1fd5){
        /* It's AT29C010A */
       /* Programming flash ROM byte by byte */
       for (i = 0; i < uLen; i+=128) {
           if (!WL_FlashWriteByte128(pAdaptor, BufAddress>>15, (BufAddress & 0x7fff), pbuf)) {
               printk("wlapi.c: WL_FlashWriteByte(buf[%d]) failed!\n", i);
               if (bRunning)
                   WL_UnHoldSutro(pAdaptor);
               return FALSE;
           }
           pbuf+=128;
           BufAddress+=128;
       }
       return TRUE;
    }
    else {
        /* It's not AT29C010A */
        printk("wlapi.c: Flash ROM type (0x%x) is unknown!\n", uFlashID);
        if (bRunning)
            WL_UnHoldSutro(pAdaptor);
        return FALSE;
    }
    /* Reset command */
/*    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xAA); */
/*    WL_FlashOutB(pAdaptor, 0, 0x2AAA, 0x55); */
/*    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xF0); */
/*    NOPLOOP(10000); */
    
/*    if (bRunning)    */
/*        WL_UnHoldSutro(pAdaptor); */
        
    return TRUE;
}

/*
 * Mask interrupt from SUTRO. (i.e. SUTRO cannot interrupt the HOST)
 * Return: TRUE if interrupt is originally enabled
 */
BOOL WL_BlockInterrupt(WL_Adaptor *pAdaptor)
{
    BYTE    cOld, cNew;
    
    /* ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_TRIVIAL_FUNC) == 0); */
    /* WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_BLK_INTR); */
    
    cOld = InB(pAdaptor->nBaseAddr + NIC_GCR);
    cNew = cOld & (~(GCR_ECINT | GCR_INT2EC | GCR_ENECINT));
    OutB(cNew, pAdaptor->nBaseAddr + NIC_GCR);
    
    /* WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_BLK_INTR); */
    return (cOld & GCR_ENECINT);
}

/*
 * Enable interrupt from SUTRO. (i.e. SUTRO can interrupt the HOST)
 * Return: TRUE if interrupt is originally enabled
 */
BOOL WL_UnBlockInterrupt(WL_Adaptor *pAdaptor)
{
    BYTE    cOld, cNew;
    
    ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_TRIVIAL_FUNC) == 0);
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_UNBLK_INTR);
    
    cOld = InB(pAdaptor->nBaseAddr + NIC_GCR);
    cNew = (cOld  & ~(GCR_ECINT | GCR_INT2EC)) | GCR_ENECINT;
    OutB(cNew, pAdaptor->nBaseAddr + NIC_GCR);
    
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_UNBLK_INTR);
    return (cOld & GCR_ENECINT);
}

/*
 * Hold SUTRO. (i.e. make SUTRO stop)
 * Return: TRUE if SUTRO is originally running
 */
BOOL WL_HoldSutro(WL_Adaptor *pAdaptor)
{
    BYTE    cOld, cNew;
    
    cOld = InB(pAdaptor->nBaseAddr + NIC_GCR);
    cNew = (cOld & ~(GCR_ECINT | GCR_INT2EC)) | GCR_ECWAIT;
    OutB(cNew, pAdaptor->nBaseAddr + NIC_GCR);

    return (cOld & GCR_ECWAIT) == 0;
}

/*
 * UnHold SUTRO. (i.e. make SUTRO running)
 * Return: TRUE if SUTRO is originally running
 */
BOOL WL_UnHoldSutro(WL_Adaptor *pAdaptor)
{
    BYTE    cOld, cNew;
    
    cOld = InB(pAdaptor->nBaseAddr + NIC_GCR);
    cNew = cOld  & (~(GCR_ECINT | GCR_INT2EC  | GCR_ECWAIT));
    OutB(cNew, pAdaptor->nBaseAddr + NIC_GCR);

    return (cOld & GCR_ECWAIT) == 0;
}

/*
 * Info card to execute a command
 */
void WL_InteruptWla(WL_Adaptor *pAdaptor)
{
    BYTE    cTmp;
    
    ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_TRIVIAL_FUNC) == 0);
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_INT_WLA);
    
    cTmp = InB(pAdaptor->nBaseAddr + NIC_GCR);
    cTmp = (cTmp | GCR_INT2EC) &  (~GCR_ECINT);
    OutB(cTmp, pAdaptor->nBaseAddr + NIC_GCR);

    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_INT_WLA);
}

/*
 *  Move 'nSize' bytes from PC to card. (Shouldn't be interrupted)
 *
 *  uDest = Card addressing space
 *  pSrc  = PC addressing space
 *  nSize = Bytes to move
 */
void WL_SetToWla(WL_Adaptor *pAdaptor, WORD uDest, void *pSrc, int nSize)
{
    ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_TRIVIAL_FUNC) == 0);
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_SET_WLA);
    
    /* switch to SRAM Page 0 */
    if (uDest & 0x8000){
        OutB(BSS_SPAGE1, pAdaptor->nBaseAddr + NIC_BSS);
    }
    else{
        OutB(BSS_SPAGE0, pAdaptor->nBaseAddr + NIC_BSS);
    }
/*    OutB(((uDest>> 11) & 0x18), pAdaptor->nBaseAddr + NIC_BSS);*/
    
    /* set LMAL and LMAH */
    OutB(uDest & 0xff, pAdaptor->nBaseAddr + NIC_LMAL);
    OutB(((uDest >> 8) & 0x7f),   pAdaptor->nBaseAddr + NIC_LMAH);
    
    /* rep out to Port A */
    OutSB(pAdaptor->nBaseAddr + NIC_IODPA, pSrc, nSize);

    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_SET_WLA);
}

/*
 *  Move 'nSize' bytes from card to PC. (Shouldn't be interrupted)
 *
 *  pSrc  = Card addressing space
 *  uDest = PC addressing space
 *  nSize = Bytes to move
 */
void WL_GetFromWla(WL_Adaptor *pAdaptor, WORD uSrc, void *pDest, int nSize)
{
    ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_TRIVIAL_FUNC) == 0);
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_GET_WLA);
    
    /* switch to SRAM Page 0 */
    if (uSrc & 0x8000){
        OutB(BSS_SPAGE1, pAdaptor->nBaseAddr + NIC_BSS);
    }
    else{
        OutB(BSS_SPAGE0, pAdaptor->nBaseAddr + NIC_BSS);
    }
/*    OutB(((uSrc>> 11) & 0x18), pAdaptor->nBaseAddr + NIC_BSS);*/
    
    /* set LMAL and LMAH */
    OutB(uSrc & 0xff, pAdaptor->nBaseAddr + NIC_LMAL);
    OutB(((uSrc >> 8) & 0x7f),   pAdaptor->nBaseAddr + NIC_LMAH);
    
    /* rep get from Port A */
    insb(pAdaptor->nBaseAddr + NIC_IODPA, pDest, nSize);

    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_GET_WLA);
}

/*
 * Get/Allocate a free Tx Data Buffer
 *
 *  *--------------*-----------------*----------------------------------*
 *  |    PLCP      |    MAC Header   |  DST  SRC         Data ...       |
 *  |  (24 bytes)  |    (30 bytes)   |  (6)  (6)  (Ethernet Row Data)   |
 *  *--------------*-----------------*----------------------------------*
 *  \               \- IEEE 802.11 -/ \------------ uDataLen -----------/
 *    \-------- TX_HEADER --------/    \-------- Ethernet Frame -------/
 *
 * Return = Postion in Card
 */
WORD WL_GetTxBuffer(WL_Adaptor *pAdaptor, WORD uDataLen)
{
    WORD  uRet,uNext,uBlockCount,uZero;
    WORD  uFullLen = sizeof(TX_HEADER) + uDataLen;
    if ( uFullLen > pAdaptor->uTxBufferCount*254 ) return 0;
    uRet= pAdaptor->uTxBufferHead;
    uBlockCount=uZero=0;
    while (uFullLen>0){
        if (uFullLen<254) uFullLen=0;
        else uFullLen-=254;
        WL_GetFromWla(pAdaptor, pAdaptor->uTxBufferHead, &uNext, sizeof(uNext));
        if (uFullLen==0){
            WL_SetToWla(pAdaptor, pAdaptor->uTxBufferHead, &uZero, sizeof(uZero));
        }
        pAdaptor->uTxBufferHead=uNext;
        uBlockCount++;
        /* if buffer is not enough */
        if ((uNext==0) && (uFullLen !=0)){
            pAdaptor->uTxBufferHead=uRet;
            return 0;
        }
    }
    pAdaptor->uTxBufferCount-=uBlockCount;
    return uRet;
}

/*
 * Free an allocated Tx Buffer. uPtr must be correct position.
 *
 */
void WL_FreeTxBuffer(WL_Adaptor *pAdaptor, WORD uPtr)
{
    WORD uNext;

    /* check if all space is not free */
    if (pAdaptor->uTxBufferHead == 0){
        pAdaptor->uTxBufferHead = uPtr;
    }
    else{
        WL_SetToWla(pAdaptor, pAdaptor->uTxBufferTail, &uPtr, sizeof(uPtr));
    }
    while(uPtr != 0){
        pAdaptor->uTxBufferCount++;
        WL_GetFromWla(pAdaptor, uPtr, &uNext, sizeof(uNext));
        pAdaptor->uTxBufferTail = uPtr;
        uPtr=uNext;
    }
    return;
}


void WL_GetMibValue(WL_Adaptor *pAdaptor, WORD cIndex)
{
    WORD uPtr;
    struct t_GetReq pGetReq;
    ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_MIB_FUNC) == 0);
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_GET_MIB);
    pGetReq.NextBlock=0;
    pGetReq.SignalID=GetRequest;
    pGetReq.MibAtrib=cIndex;
    if (WL_ESBQRequestTest(pAdaptor)){
        uPtr= WL_GetTxBuffer(pAdaptor, sizeof(pGetReq));
        if (uPtr !=0){
            WL_SetToWla(pAdaptor, uPtr, &pGetReq, sizeof(pGetReq));
            WL_ESBQRequest(pAdaptor,&uPtr);
        }
    }
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_GET_MIB);
    return;
}

void WL_SetMibValue(WL_Adaptor *pAdaptor, WORD cIndex, void *pSrc, BYTE cSize)
{

    WORD uPtr;
    struct t_SetReq pSetReq;
    ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_MIB_FUNC) == 0);
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_SET_MIB);
    pSetReq.NextBlock=0;
    pSetReq.SignalID=SetRequest;
    pSetReq.MibAtrib=cIndex;
    memcpy((char *) &pSetReq.MibValue,(char *) pSrc,cSize);
    if (WL_ESBQRequestTest(pAdaptor)){
        uPtr= WL_GetTxBuffer(pAdaptor, sizeof(pSetReq));
        if (uPtr !=0){
            WL_SetToWla(pAdaptor, uPtr, &pSetReq, sizeof(pSetReq));
            WL_ESBQRequest(pAdaptor,&uPtr);
        }
    }
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_SET_MIB);
    return;
}

BOOL WL_ResetBoard(WL_Adaptor *pAdaptor)
{
    BYTE    cTmp;
    DWORD   i;
    
    /* Coreset */
    OutB_P(GCR_CORESET, pAdaptor->nBaseAddr + NIC_GCR);
    NOPLOOP(5000);
    OutB_P(0,           pAdaptor->nBaseAddr + NIC_GCR);
    NOPLOOP(5000);
    OutB_P(GCR_CORESET, pAdaptor->nBaseAddr + NIC_GCR);

    /* Reset SRAM 0x480 to zero */
    cTmp = 0;
    OutB(BSS_SPAGE0, pAdaptor->nBaseAddr + NIC_BSS);
    WL_SetToWla(pAdaptor, 0x480, &cTmp, sizeof(cTmp));
    
    /* Start up */
    OutB_P(0, pAdaptor->nBaseAddr + NIC_GCR);
    
    /* NOP */
    NOPLOOP(1024*50);

    /* WL_UnBlockInterrupt(pAdaptor); */

    /* Polling Self_Test_Status */    
    for (i = 0; i < 1000; i++) {
        WL_GetFromWla(pAdaptor, 0x480, &cTmp, sizeof(cTmp));

        if (cTmp == 'W') {
            /* firmware complete all test successfully */
            cTmp = 'A';
            WL_SetToWla(pAdaptor, 0x480, &cTmp, sizeof(cTmp));
            return TRUE;
        }
        NOPLOOP(5000);
    }

    return FALSE;
}

BOOL WL_InitFirmware(WL_Adaptor *pAdaptor)
{
    WORD    uPtr,uNext;
    
    ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_API_FUNC) == 0);
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_INIT);
    
    if (!WL_ResetBoard(pAdaptor)) {
        TRACE_INIT("wlapi.c: WL_ResetBoard failed!\n");
        WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_INIT);
        return FALSE;
    }
        
    /* Switch to SRAM Page 0 */
    OutB(BSS_SPAGE0, pAdaptor->nBaseAddr + NIC_BSS);
    /* Read parameter from card */
    WL_GetFromWla(pAdaptor, 0x482, &pAdaptor->uESBQReqStart,2);
    WL_GetFromWla(pAdaptor, 0x486, &pAdaptor->uESBQReqEnd,2);
    WL_GetFromWla(pAdaptor, 0x488, &pAdaptor->uESBQConfirmStart,2);
    WL_GetFromWla(pAdaptor, 0x48C, &pAdaptor->uESBQConfirmEnd,2);
    WL_GetFromWla(pAdaptor, 0x48E, &pAdaptor->uTxBufferHead,2);
    WL_GetFromWla(pAdaptor, 0x492, &pAdaptor->uTxBufferSize,2);
    pAdaptor->uESBQReqTail=pAdaptor->uESBQReqHead=pAdaptor->uESBQReqStart;
    pAdaptor->uESBQReqEnd += pAdaptor->uESBQReqStart;
    pAdaptor->uESBQConfirm = pAdaptor->uESBQConfirmStart;
    pAdaptor->uESBQConfirmEnd += pAdaptor->uESBQConfirmStart;
    /* Initial Tx Buffer */
    pAdaptor->uTxBufferCount=1;
    uPtr = pAdaptor->uTxBufferHead;
    uNext = uPtr + BlockSize;
    while ( (uNext- pAdaptor->uTxBufferHead) < pAdaptor->uTxBufferSize){
        pAdaptor->uTxBufferCount++;
        WL_SetToWla(pAdaptor, uPtr, &uNext, sizeof(uNext));
        uPtr = uNext;
        uNext = uPtr + BlockSize;
    }
    uNext = 0 ;
    WL_SetToWla(pAdaptor, uPtr, &uNext, sizeof(uNext));
    pAdaptor->uTxBufferTail=uPtr;
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_INIT);
    return TRUE;    
}

/*
 * Send a packet.
 *
 * pData = Ethernet raw frame.  (e.g. pData[0] - pData[5] is Dest MAC Addr, 
 *                                    pData[6] - pData[11] is Src MAC Addr)
 * Ref: IEEE 802.11
 */
BOOL WL_SendPkt(WL_Adaptor *pAdaptor, BYTE *pData, WORD uLen)
{
    WORD    uBuf,uSig,uNext,templen,packet_length,cTemp;
    BYTE    *DataPtr;
    struct t_MdReq Signal;
    DataPtr=(char *) pData;
    ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_API_FUNC) == 0);
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_TX_PKT);

    if (WL_ESBQRequestTest(pAdaptor)){
        uSig = WL_GetTxBuffer(pAdaptor, sizeof(Signal));
        if (uSig == 0) {   /* No free buffer available */
            WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_TX_PKT);
            return FALSE;
        }

        uBuf = WL_GetTxBuffer(pAdaptor, uLen+26+24);

        if (uBuf == 0)  {
            /* No free buffer available */
            WL_FreeTxBuffer(pAdaptor, uSig);
            WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_TX_PKT);
            return FALSE;
        }
        memcpy((char *) &(Signal.DAddress[0]),DataPtr,12);
        packet_length=uLen-12;
        DataPtr +=12;
        Signal.NextBlock=0;
        Signal.SignalID=MdRequest;
        Signal.Data=uBuf;
        if (pAdaptor->LLCType==1){
            if (((*DataPtr)*256+(*(DataPtr+1)))>1500){
                Signal.Size=packet_length+24+4+6;
                cTemp=0xaaaa;
                WL_SetToWla(pAdaptor, uBuf+2+sizeof(TxHeader)-6, &cTemp, sizeof(cTemp));
                cTemp=0x03;
                WL_SetToWla(pAdaptor, uBuf+2+sizeof(TxHeader)-4, &cTemp, sizeof(cTemp));
                cTemp=0x0;
                WL_SetToWla(pAdaptor, uBuf+2+sizeof(TxHeader)-2, &cTemp, sizeof(cTemp));
                if (packet_length>(254-sizeof(TxHeader))){
                    templen=(254-sizeof(TxHeader));
                    packet_length-=templen;
                }
                else{
                    templen=packet_length;
                    packet_length=0;
                }
                WL_SetToWla(pAdaptor, uBuf+2+sizeof(TxHeader), DataPtr, templen);
                DataPtr+=templen;
                WL_GetFromWla(pAdaptor, uBuf, &uNext, sizeof(uNext));
                uBuf=uNext;
            }
            else{
                Signal.Size=packet_length+24+4-2;
                DataPtr+=2;
                packet_length-=2;
                if (packet_length>(254-sizeof(TxHeader)+6)){
                    templen=(254-sizeof(TxHeader)+6);
                    packet_length-=templen;
                }
                else{
                    templen=packet_length;
                    packet_length=0;
                }
                WL_SetToWla(pAdaptor, uBuf+2+sizeof(TxHeader)-6, DataPtr, templen);
                DataPtr+=templen;
                WL_GetFromWla(pAdaptor, uBuf, &uNext, sizeof(uNext));
                uBuf=uNext;
            }
        }
        else{
            packet_length+=12;
            Signal.Size=packet_length+24+4;
            DataPtr-=12;
            if (packet_length>(254-sizeof(TxHeader)+6)){
                templen=(254-sizeof(TxHeader)+6);
                packet_length-=templen;
            }
            else{
                templen=packet_length;
                packet_length=0;
            }
            WL_SetToWla(pAdaptor, uBuf+2+sizeof(TxHeader)-6, DataPtr, templen);
            DataPtr+=templen;
            WL_GetFromWla(pAdaptor, uBuf, &uNext, sizeof(uNext));
            uBuf=uNext;
        }
        while (packet_length >0){
            if (packet_length>254){
                templen=254;
                packet_length-=254;
            }
            else{
                templen=packet_length;
                packet_length=0;
            }
            WL_SetToWla(pAdaptor, uBuf+2, DataPtr, templen);
            DataPtr+=templen;
            WL_GetFromWla(pAdaptor, uBuf, &uNext, sizeof(uNext));
            uBuf=uNext;
        }
        WL_SetToWla(pAdaptor, uSig, &Signal, sizeof(MdReq));
        WL_ESBQRequest(pAdaptor,&uSig);
    }
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_TX_PKT);
    return TRUE;    
}


/*
 * Receive data from Receive Queue.
 *
 * pBuf = address of host 
 * uSize = size of buffer.
 *
 * Scenario:
 *
 * Return: 0
 */
WORD WL_Receive(WL_Adaptor *pAdaptor,BYTE *pBuf, WORD uSize)
{
    WORD    NextAddress,NextAddress1;
    BYTE    *pData;
    pData=pBuf+12;
    uSize-=12;
    WL_GetFromWla(pAdaptor, pAdaptor->StartSegment+2, &NextAddress, sizeof(NextAddress));
    if (pAdaptor->LLCType==1){
        if (pAdaptor->EtherType==EtherII){
            if (uSize>BlockSize-sizeof(RxHeader)){
                WL_GetFromWla(pAdaptor, pAdaptor->StartSegment+sizeof(RxHeader), pData, BlockSize-sizeof(RxHeader));
                uSize=uSize-(BlockSize-sizeof(RxHeader));
                pData=pData+(BlockSize-sizeof(RxHeader));
            }
            else{
                WL_GetFromWla(pAdaptor, pAdaptor->StartSegment+sizeof(RxHeader), pData, uSize);
                uSize=0;
            }
        }
        else{
            uSize-=2;
            *pData=(uSize>>8) & 0xff;
            *(pData+1)=uSize & 0xff;
            pData+=2;
            if (uSize>BlockSize-sizeof(RxHeader)+6){
                WL_GetFromWla(pAdaptor, pAdaptor->StartSegment+sizeof(RxHeader)-6, pData, BlockSize-sizeof(RxHeader)+6);
                uSize=uSize-(BlockSize-sizeof(RxHeader))+6;
                pData=pData+(BlockSize-sizeof(RxHeader)+6);
            }
            else{
                WL_GetFromWla(pAdaptor, pAdaptor->StartSegment+sizeof(RxHeader)-6, pData, uSize);
                uSize=0;
            }
        }
    }
    else{
            if (uSize>BlockSize-sizeof(RxHeader)-6){
                WL_GetFromWla(pAdaptor, pAdaptor->StartSegment+sizeof(RxHeader)+6, pData, BlockSize-sizeof(RxHeader)-6);
                uSize=uSize-(BlockSize-sizeof(RxHeader))-6;
                pData=pData+(BlockSize-sizeof(RxHeader)-6);
            }
            else{
                WL_GetFromWla(pAdaptor, pAdaptor->StartSegment+sizeof(RxHeader)+6, pData, uSize);
                uSize=0;
            }

    }
    while (uSize>0){
        if (uSize>BlockSize-5){
           WL_GetFromWla(pAdaptor, NextAddress+5, pData, BlockSize-5);
           uSize=uSize-(BlockSize-5);
           pData=pData+(BlockSize-5);
           WL_GetFromWla(pAdaptor, NextAddress+2, &NextAddress1, sizeof(NextAddress1));
           NextAddress=NextAddress1;
        }
        else{
           WL_GetFromWla(pAdaptor, NextAddress+5, pData, uSize);
           uSize=0;
        }
    }
    return 0;
}


BOOL WL_ESBQRequestTest(WL_Adaptor *pAdaptor)
{
    BYTE cTmp;
    WL_GetFromWla(pAdaptor, pAdaptor->uESBQReqHead+3, &cTmp, sizeof(cTmp));
    if ((cTmp & 0x80) == 0) return FALSE;
    return TRUE;
}
void WL_ESBQRequest(WL_Adaptor *pAdaptor, WORD *ptr)
{
    WORD cTmp=0x0;
    WL_SetToWla(pAdaptor, pAdaptor->uESBQReqHead, ptr, 2);
    WL_SetToWla(pAdaptor, pAdaptor->uESBQReqHead+2, &cTmp, sizeof(cTmp));
    pAdaptor->uESBQReqHead+=4;
    if (pAdaptor->uESBQReqHead>=pAdaptor->uESBQReqEnd)
        pAdaptor->uESBQReqHead=pAdaptor->uESBQReqStart;
    return;
}
void WL_ESBQRequestFree(WL_Adaptor *pAdaptor)
{
    BYTE cTmp;
    WORD Address;
    if (pAdaptor->uESBQReqHead != pAdaptor->uESBQReqTail){
        WL_GetFromWla(pAdaptor, pAdaptor->uESBQReqTail+3, &cTmp, sizeof(cTmp));
        if ((cTmp & 0x80) !=0){
            WL_GetFromWla(pAdaptor, pAdaptor->uESBQReqTail, &Address, sizeof(Address));
            WL_FreeTxBuffer(pAdaptor,Address);
            pAdaptor->uESBQReqTail+=4;
            if (pAdaptor->uESBQReqTail>=pAdaptor->uESBQReqEnd)
                pAdaptor->uESBQReqTail=pAdaptor->uESBQReqStart;
        }
    }
    return;
}
BOOL WL_ESBQConfirm(WL_Adaptor *pAdaptor)
{
    BYTE cTmp;
    WL_GetFromWla(pAdaptor, pAdaptor->uESBQConfirm+3, &cTmp, sizeof(cTmp));
    if ((cTmp & 0x80) == 0) return FALSE;
    return TRUE;
}
void WL_ESBQConfirmDone(WL_Adaptor *pAdaptor)
{
    BYTE cTmp=0x0;
    WL_SetToWla(pAdaptor, pAdaptor->uESBQConfirm+3, &cTmp, sizeof(cTmp));
    pAdaptor->uESBQConfirm+=4;
    if (pAdaptor->uESBQConfirm>=pAdaptor->uESBQConfirmEnd)
        pAdaptor->uESBQConfirm=pAdaptor->uESBQConfirmStart;
    return;
}
BOOL MgmtStart(WL_Adaptor *pAdaptor){
    struct t_StartReq Signal;
    WORD uPtr;
    Signal.NextBlock=0;
    Signal.SignalID=StartRequest;
    memcpy((char *) Signal.SSID,(char *) pAdaptor->ESSID,34);
    memcpy((char *) pAdaptor->KeepESSID,(char *) pAdaptor->ESSID,34);
    Signal.BSSType=Independent;
    Signal.BeaconPeriod=400;
    Signal.DTIMPeriod=1;
    Signal.PHYpset[0]=3;
    Signal.PHYpset[1]=1;
    Signal.PHYpset[2]=pAdaptor->DefaultChannel;
    Signal.CapablilityInfo=0x02;
    Signal.BSSBasicRateSet[0]=0x01;
    Signal.BSSBasicRateSet[1]=0x02;
    Signal.BSSBasicRateSet[2]=0x82;
    Signal.BSSBasicRateSet[3]=0x84;
    Signal.OperationalRateSet[0]=0x01;
    Signal.OperationalRateSet[1]=0x02;
    Signal.OperationalRateSet[2]=0x82;
    Signal.OperationalRateSet[3]=0x84;
    Signal.IBSSpset[0]=6;
    Signal.IBSSpset[1]=2;
    Signal.IBSSpset[2]=10;
    Signal.IBSSpset[3]=0;
    if (WL_ESBQRequestTest(pAdaptor)){
        uPtr= WL_GetTxBuffer(pAdaptor, sizeof(Signal));
        if (uPtr !=0){
            WL_SetToWla(pAdaptor, uPtr, &Signal, sizeof(Signal));
            WL_ESBQRequest(pAdaptor,&uPtr);
            return TRUE;
        }
    }
    return FALSE;
}
BOOL MgmtScan(WL_Adaptor *pAdaptor,WORD ChannelTime){
    struct t_ScanReq Signal;
    WORD uPtr;

    Signal.NextBlock=0;
    Signal.SignalID=ScanRequest;
    Signal.SSID[0]=0;
    Signal.SSID[1]=0;
    Signal.ScanType=ActiveScan;
    Signal.ProbeDelay=0x10;
    Signal.MinChannelTime=ChannelTime;
    Signal.MaxChannelTime=ChannelTime;
    if (pAdaptor->NetworkType == Infrastructure){
        Signal.BSSType=Infrastructure;
    }
    else{
        Signal.BSSType=Independent;
    }
    pAdaptor->BSSCount=0;
    pAdaptor->JoinStaBSS=0;
    if (WL_ESBQRequestTest(pAdaptor)){
        uPtr= WL_GetTxBuffer(pAdaptor, sizeof(Signal));
        if (uPtr !=0){
            WL_SetToWla(pAdaptor, uPtr, &Signal, sizeof(Signal));
            WL_ESBQRequest(pAdaptor,&uPtr);
            return TRUE;
        }
    }
    return FALSE;
}
BOOL MgmtJoin(WL_Adaptor *pAdaptor,WORD stas){
    struct t_JoinReq Signal;
    WORD uPtr;

    Signal.NextBlock=0;
    Signal.SignalID=JoinRequest;
    Signal.Timeout=10;
    memcpy((char *) &(Signal.BeaconPeriod),(char *) &(pAdaptor->BSSset[stas].BeaconPeriod),72);
    pAdaptor->CapablilityInfo=Signal.CapablilityInfo;
    pAdaptor->Channel=Signal.PHYpset[2];
    if (WL_ESBQRequestTest(pAdaptor)){
        uPtr= WL_GetTxBuffer(pAdaptor, sizeof(Signal));
        if (uPtr !=0){
            WL_SetToWla(pAdaptor, uPtr, &Signal, sizeof(Signal));
            WL_ESBQRequest(pAdaptor,&uPtr);
            return TRUE;
        }
    }
    return FALSE;
}
BOOL MgmtAuthetication(WL_Adaptor *pAdaptor){
    struct t_AutheReq Signal;
    WORD uPtr;

    Signal.NextBlock=0;
    Signal.SignalID=AutheRequest;
    Signal.Type=Open_System;
    Signal.Timeout=1000;
    memcpy((char *) &(Signal.MacAddress),(char *) &(pAdaptor->BSSID),6);
    if (WL_ESBQRequestTest(pAdaptor)){
        uPtr= WL_GetTxBuffer(pAdaptor, sizeof(Signal));
        if (uPtr !=0){
            WL_SetToWla(pAdaptor, uPtr, &Signal, sizeof(Signal));
            WL_ESBQRequest(pAdaptor,&uPtr);
            return TRUE;
        }
    }
    return FALSE;
}
BOOL MgmtAssociation(WL_Adaptor *pAdaptor){
    struct t_AssocReq Signal;
    WORD uPtr;

    Signal.NextBlock=0;
    Signal.SignalID=AssocRequest;
    Signal.Timeout=1000;
    Signal.ListenInterval=5;
    Signal.CapInfo=pAdaptor->CapablilityInfo;
    memcpy((char *) &(Signal.MacAddress),(char *) &(pAdaptor->BSSID),6);
    if (WL_ESBQRequestTest(pAdaptor)){
        uPtr= WL_GetTxBuffer(pAdaptor, sizeof(Signal));
        if (uPtr !=0){
            WL_SetToWla(pAdaptor, uPtr, &Signal, sizeof(Signal));
            WL_ESBQRequest(pAdaptor,&uPtr);
            return TRUE;
        }
    }
    return FALSE;
}
BOOL MgmtDeauthetication(WL_Adaptor *pAdaptor){
    struct t_DeautheReq Signal;
    WORD uPtr;

    Signal.NextBlock=0;
    Signal.SignalID=DeautheRequest;
    Signal.Reason=8;
    memcpy((char *) &(Signal.MacAddress),(char *) &(pAdaptor->BSSID),6);
    if (WL_ESBQRequestTest(pAdaptor)){
        uPtr= WL_GetTxBuffer(pAdaptor, sizeof(Signal));
        if (uPtr !=0){
            WL_SetToWla(pAdaptor, uPtr, &Signal, sizeof(Signal));
            WL_ESBQRequest(pAdaptor,&uPtr);
            return TRUE;
        }
    }
    return FALSE;
}
BOOL MgmtResync(WL_Adaptor *pAdaptor){
    struct t_ResyncReq Signal;
    WORD uPtr;

    Signal.NextBlock=0;
    Signal.SignalID=ResyncRequest;
    if (WL_ESBQRequestTest(pAdaptor)){
        uPtr= WL_GetTxBuffer(pAdaptor, sizeof(Signal));
        if (uPtr !=0){
            WL_SetToWla(pAdaptor, uPtr, &Signal, sizeof(Signal));
            WL_ESBQRequest(pAdaptor,&uPtr);
            return TRUE;
        }
    }
    return FALSE;
}
void MgmtScanConfirm(WL_Adaptor *pAdaptor,WORD Address){
    WORD i,j;
    BOOL matchflag;
    struct t_ScanConf Signal;
    matchflag=FALSE;
    WL_GetFromWla(pAdaptor, Address, &Signal, sizeof(Signal));
    if (Signal.Status == SUCCESS){
        if (pAdaptor->DriverState == ScanRequest){
            for (i=0;i<pAdaptor->BSSCount;i++){
                if (memcmp((char *) pAdaptor->BSSset[i].BSSID,(char *) Signal.BSSID,6)==0)
                    break;
            }
            if ((i == pAdaptor->BSSCount) && (i<20) ){
                memcpy((char *) &(pAdaptor->BSSset[i].BeaconPeriod),(char *) &(Signal.BeaconPeriod),73);
                pAdaptor->BSSCount++;
            }
        }
        else if ( ((pAdaptor->NetworkType == Infrastructure) && ((Signal.CapablilityInfo & 0x01) != 0))
            || ((pAdaptor->NetworkType == Independent) && ((Signal.CapablilityInfo & 0x02) != 0))
            || (pAdaptor->NetworkType == 2) ){
                if (pAdaptor->ESSID[1]==0) matchflag=TRUE;
		else if ((pAdaptor->ESSID[1]==3) && (strncmp((char *) &(pAdaptor->ESSID[2]),"ANY", 3) == 0)){
			matchflag=TRUE;
		}
                else if (pAdaptor->ESSID[1] != Signal.SSID[1]) matchflag=FALSE;
                else if (memcmp((char *) &(pAdaptor->ESSID[2]),(char *) &(Signal.SSID[2]),pAdaptor->ESSID[1]))
                    matchflag=FALSE;
                else matchflag=TRUE;
                if (matchflag){
                    for (i=0;i<pAdaptor->BSSCount;i++){
                        if (memcmp((char *) pAdaptor->BSSset[i].BSSID,(char *) Signal.BSSID,6) ==0){
                            matchflag=FALSE;
                            break;
                        }
                    }
                }
                if (matchflag && (i<20)){
                    memcpy((char *) &(pAdaptor->BSSset[i].BeaconPeriod),(char *) &(Signal.BeaconPeriod),73);
                    pAdaptor->BSSCount++;
                }
        }
    }
    else if ((Signal.Status == TIMEOUT) && (pAdaptor->DriverState != ScanRequest)){
        pAdaptor->JoinStaBSS=0;
        for (j=pAdaptor->JoinStaBSS;j<pAdaptor->BSSCount;j++){
            for (i=0;i<3000;i++){
                if (MgmtJoin(pAdaptor,j)) break;
            }
            if (i != 3000) break;
        }
        pAdaptor->JoinStaBSS=j;
        if (pAdaptor->JoinStaBSS == pAdaptor->BSSCount){
            if (pAdaptor->NetworkType == Infrastructure){
                for (i=0;i<3000;i++){
                    if (MgmtScan(pAdaptor,100)) break;
                }
            }
            else{
                pAdaptor->AdHocTimes++;
                if (pAdaptor->AdHocTimes>8){
                    for (i=0;i<3000;i++){
                        if (MgmtStart(pAdaptor)) break;
                    }
                }
                else{
                    for (i=0;i<3000;i++){
                        if (MgmtScan(pAdaptor,100)) break;
                    }
                }
            }

        }

    }

}
void MgmtJoinConfirm(WL_Adaptor *pAdaptor,WORD Address){
    WORD i,j;
    struct t_JoinConf Signal;

    WL_GetFromWla(pAdaptor, Address, &Signal, sizeof(Signal));
    if (Signal.Status == SUCCESS){
        if (pAdaptor->NetworkType == Infrastructure){
            if (pAdaptor->JoinStaBSS < pAdaptor->BSSCount){
                i=pAdaptor->JoinStaBSS;
                memcpy((char *) &(pAdaptor->BSSID),(char *) &(pAdaptor->BSSset[i].BSSID),6);
                pAdaptor->Channel=pAdaptor->BSSset[i].PHYpset[2];
                memcpy((char *) pAdaptor->KeepESSID,(char *) pAdaptor->BSSset[i].SSID,34);
                for (i=0;i<3000;i++){
                    if (MgmtAuthetication(pAdaptor)) break;
                }
            }
        }
        else{
                pAdaptor->CardStart=TRUE;
                i=pAdaptor->JoinStaBSS;
                memcpy((char *) &(pAdaptor->BSSID),(char *) &(pAdaptor->BSSset[i].BSSID),6);
                pAdaptor->Channel=pAdaptor->BSSset[i].PHYpset[2];
                memcpy((char *) pAdaptor->KeepESSID,(char *) pAdaptor->BSSset[i].SSID,34);

                        printf("Join BSS: "); 
			printf("%02x:%02x:%02x:%02x:%02x:%02x, ESSID: %s, Channel: %d\n",
				pAdaptor->BSSID.b0,pAdaptor->BSSID.b1,
				pAdaptor->BSSID.b2,pAdaptor->BSSID.b3,
				pAdaptor->BSSID.b4,pAdaptor->BSSID.b5,
				twl_ESSID(pAdaptor->KeepESSID), pAdaptor->Channel);
/*
                printk("Wireless LAN On-Line  Join:");
                printk(" %2X %2X %2X %2X %2X %2X\n",pAdaptor->BSSID.b0,pAdaptor->BSSID.b1,pAdaptor->BSSID.b2,pAdaptor->BSSID.b3,pAdaptor->BSSID.b4,pAdaptor->BSSID.b5);
*/
        }
    }
    else{
        pAdaptor->JoinStaBSS++;
        for (j=pAdaptor->JoinStaBSS;j<pAdaptor->BSSCount;j++){
            for (i=0;i<3000;i++){
                if (MgmtJoin(pAdaptor,j)) break;
            }
            if (i != 3000) break;
        }
        pAdaptor->JoinStaBSS=j;
        if (pAdaptor->JoinStaBSS == pAdaptor->BSSCount){
            if (pAdaptor->NetworkType == Infrastructure){
                for (i=0;i<3000;i++){
                    if (MgmtScan(pAdaptor,100)) break;
                }
            }
            else{
                pAdaptor->AdHocTimes++;
                if (pAdaptor->AdHocTimes>8){
                    for (i=0;i<3000;i++){
                        if (MgmtStart(pAdaptor)) break;
                    }
                }
                else{
                    for (i=0;i<3000;i++){
                        if (MgmtScan(pAdaptor,100)) break;
                    }
                }
            }

        }

    }
}
void MgmtResyncConfirm(WL_Adaptor *pAdaptor){
    WORD i,j;
    pAdaptor->JoinStaBSS++;
    for (j=pAdaptor->JoinStaBSS;j<pAdaptor->BSSCount;j++){
        for (i=0;i<3000;i++){
            if (MgmtJoin(pAdaptor,j)) break;
        }
        if (i != 3000) break;
    }
    pAdaptor->JoinStaBSS=j;
    if (pAdaptor->JoinStaBSS >= pAdaptor->BSSCount){
        if (pAdaptor->NetworkType == Infrastructure){
            for (i=0;i<3000;i++){
                if (MgmtScan(pAdaptor,100)) break;
            }
        }
        else{
            pAdaptor->AdHocTimes++;
            if (pAdaptor->AdHocTimes>8){
                for (i=0;i<3000;i++){
                    if (MgmtStart(pAdaptor)) break;
                }
            }
            else{
                for (i=0;i<3000;i++){
                    if (MgmtScan(pAdaptor,100)) break;
                }
            }
        }

    }

}
/*
 * Refer AMD Am29F010 CMOS Sector Erase Flash Memory Data Sheet.
 */


/* #include <asm/segment.h>*/

/* Hack some linux functions */
#ifdef __FreeBSD__
   extern const char g_cpszVersion[];
   struct zwl_softc;
   extern int zwl_reset(struct zwl_softc *sc);
#  define WL_Reset(sc)      zwl_reset(sc)
#  define memcpy_tofs(pDest, pSrc, ulLen)       copyout(pSrc, pDest, ulLen)
#  define memcpy_fromfs(pDest, pSrc, ulLen)     copyin(pSrc, pDest, ulLen)
#  define suser()   (suser(curproc->p_ucred, &curproc->p_acflag) == 0)
#  include <sys/kernel.h>       /* for ticks */
#  define WL_DELTA_TIME(x)      ((x) = ticks - (x))

#  define VERIFY_READ   0
#  define VERIFY_WRITE  1

static int verify_area(int nOperation, const void *pData, WORD ulLen)
{
    /* always success */
    return 0;
}

#else /* For Linux */

#  include "wl24.h"
/* extern const char g_cpszVersion[]; */
/* extern int WL_Reset(struct device *dev); */

#  define WL_DELTA_TIME(x)      ((x) = jiffies - (x))

/*
 * Perform IOCTL call functions here. Some are privileged operations and the
 * effective uid is checked in those cases.
 *
 * This part is optional. Needed only if you want to run wlu (unix version).
 *
 * CAUTION: To prevent interrupted by WL_Interrupt() and timer-based 
 * WL_StartXmit() from other interrupts, this should be run single-threaded.
 * This function is expected to be a rare operation, and it's 
 * simpler to just use cli() to disable ALL interrupts.
 */
int WL_Ioctl(struct device *dev, struct ifreq *rq, int cmd)
{
    WL_Adaptor      *pAdaptor = (WL_Adaptor *) dev->priv;
    WL_IOCTL_BLK    *pBlk = (WL_IOCTL_BLK *) &rq->ifr_data;
    unsigned long   ulFlags;
    int             nReturn;

    /* Keep flags first, these are inline functions */
    save_flags(ulFlags);
    cli();

    nReturn = WL_IoctlWla(pAdaptor, pBlk, dev);
    
    /* Restore CLI and others */
    restore_flags(ulFlags);

    return nReturn;
}

#endif /* Linux End */

/*
 * Common ioctl part for both Linux and FreeBSD
 * Return 0 if success.
 */
int WL_IoctlWla(WL_Adaptor *pAdaptor, WL_IOCTL_BLK *pBlk, void *pDevice)
{
    int  nReturn;
    WL_IOCTL_PARA lParam;
    unsigned char buff[1028];
    switch(pBlk->uCmd) {
      case WL_IOCTL_SET_RESET:      /* Reset driver (needed after set) */
        if (!suser()) {
            nReturn = -EPERM;
            break;
        }
        
        nReturn = WL_Reset(pDevice);
        break;

      case WL_IOCTL_WRITE_FLASH:      /* Write firmware into Flash ROM */
        if (!suser()) {
            nReturn = -EPERM;
            break;
        }

        memcpy_fromfs( buff, pBlk->pData, pBlk->uLen);
        nReturn = WL_WriteFlash(pAdaptor,buff, pBlk->uLen) ? 0 : -EIO;
        break;

      case WL_IOCTL_GET_PARAMETER: {  /* Get parameter */
        lParam.DefaultChannel = pAdaptor->DefaultChannel;
        lParam.Channel = pAdaptor->Channel;
        lParam.NetworkType = pAdaptor->NetworkType;
        lParam.LLCType = pAdaptor->LLCType;
        lParam.Version[0] = pAdaptor->Version[0];
        lParam.Version[1] = pAdaptor->Version[1];
        lParam.FrequencyDomain = pAdaptor->FrequencyDomain;
        memcpy((char *) &(lParam.KeepESSID[0]),(char *) &(pAdaptor->KeepESSID[0]), 34);
        memcpy((char *) &(lParam.ESSID[0]), (char *) &(pAdaptor->ESSID[0]), 34);
        pBlk->uLen = sizeof(lParam);
        memcpy_tofs(pBlk->pData, &lParam, pBlk->uLen);
        return 0;
      }

      case WL_IOCTL_SET_PARAMETER: {  /* Set parameter */
        if (!suser()) {
            nReturn = -EPERM;
            break;
        }
        nReturn = verify_area(VERIFY_READ, (void *)pBlk->pData, sizeof(lParam));

        if (nReturn != 0)
            break;

        memcpy_fromfs( &lParam, pBlk->pData, sizeof(lParam));
        pAdaptor->DefaultChannel=lParam.DefaultChannel;
        pAdaptor->NetworkType=lParam.NetworkType;
        pAdaptor->LLCType=lParam.LLCType;
        memcpy((char *) &(pAdaptor->ESSID[0]), (char *) &(lParam.ESSID[0]), 34); 
        nReturn = WL_Reset(pDevice);
        break;
      }
      default:
    	nReturn = -EOPNOTSUPP;
    }
    
    return nReturn;
}

/*
 * ---------------------------------------------------------------------------
 * Plug'n'play routines
 * ---------------------------------------------------------------------------
 */


static unsigned char INIT_KEY_TABLE[34]={
    0x00,0x00,
    0x6A,0xB5,0xDA,0xED,0xF6,0xFB,0x7D,0xBE,
    0xDF,0x6F,0x37,0x1B,0x0D,0x86,0xC3,0x61,
    0xB0,0x58,0x2C,0x16,0x8B,0x45,0xA2,0xD1,
    0xE8,0x74,0x3A,0x9D,0xCE,0xE7,0x73,0x39
};

BOOL CompareCheckSum(unsigned char Data[])
{
    unsigned char i;
    unsigned char checksum = 0x6A;

    for (i = 0; i < 8; i++) {
        unsigned char bit;
        unsigned char ch;

        ch = Data[i];
        for (bit = 0; bit < 8; bit++) {
             unsigned char temp = (checksum^(checksum>>1)^ch) & 0x01;
             checksum >>= 1;
             ch >>= 1;
             checksum += temp << 7;
        }
    }
        
    return (checksum == Data[8]);
}

void GetResourceData(unsigned char ResourceData[], int nPort,int nCSN)
{
    int i;

    /* Wake [CSN] */
    OutB(0x03, PNP_ADDR_PORT);
    OutB(nCSN, PNP_WR_PORT);

    /* set read data port */
    OutB(0x00, PNP_ADDR_PORT);
    OutB(nPort>>2, PNP_WR_PORT);

    /* Read Resource Data */
    OutB(0x04, PNP_ADDR_PORT);
    NOPLOOP(100);

    for (i=0; i<9 ; i++) {
        NOPLOOP(100);
        ResourceData[i] = InB(nPort);
    }
}

void GetSerialID(unsigned char SerialID[], int nPnPRdPort)
{
    int i,j;
    unsigned char temp1,temp2;

    /* Wake [0] */
    OutB(0x03, PNP_ADDR_PORT);
    OutB(0x00, PNP_WR_PORT);

    /* Set Read Data Port */
    OutB(0x00, PNP_ADDR_PORT);
    OutB(nPnPRdPort>>2, PNP_WR_PORT);

    /* Read 72 pair from serial isolation port */
    OutB(0x01, PNP_ADDR_PORT);

    for (i=0; i<9; i++) {
        SerialID[i] = 0;        /* by fox */
        for (j=0; j<8 ; j++) {
            SerialID[i] >>= 1 ;

            temp1 = InB(nPnPRdPort);
            NOPLOOP(100);

            temp2 = InB(nPnPRdPort);
            if (temp1==0x55 && temp2==0xAA) SerialID[i]|=0x80;
            NOPLOOP(100);
        }
    }
}

/*
 * Return PnP DataPort, or 0 if failed
 */
int Find_ReadDataPort(void)
{
    unsigned char buf[9];
    int nPort;

    for (nPort = 0x203; nPort < 0x300; nPort += 4) {
        GetResourceData(buf, nPort, 1);
        if (CompareCheckSum(buf)) {
            printk("wlpnp.c: Found Data port = 0x%3.3x, by resource data\n", nPort);
            return nPort;
        }
    }

    for (nPort = 0x203; nPort <= 0x300; nPort += 4) {
        GetSerialID(buf,nPort);
        if (CompareCheckSum(buf)) {
            printk("wlpnp.c: Found Data port = 0x%3.3x, by serial ID\n", nPort);
            return nPort;
        }
    }

    return 0;
}

/*
 * Return Total CSN found
 */
int Find_TotalCSN(int nPort)
{
    unsigned char buf[9];
    int i;

    for (i = 1; i < DEV_NUM_MAX; i++) {
        GetResourceData(buf, nPort, i);
        if (!CompareCheckSum(buf)) {
            printk("wlpnp.c: Total %d CSN found\n", i-1);
            return i-1;
        }
    }
    return i;
}

void WL_GetDeviceInfo(int CSN,int nPnPReadPort, DWORD *pulBaseAddr, BYTE *pcIRQ)
{
    unsigned int iobase, irq;
    
    /* Wake [CSN] */
    OutB(0x03,  PNP_ADDR_PORT);
    OutB(CSN,   PNP_WR_PORT);
    OutB(0x60,  PNP_ADDR_PORT);    /* IO Port Base [15:8] */
    iobase = (int) InB(nPnPReadPort);
    iobase <<= 8;
    OutB(0x61, PNP_ADDR_PORT);    /* IO Port Base [7:0] */
    iobase += InB(nPnPReadPort);
    OutB(0x70, PNP_ADDR_PORT);    /* IRQ Level */
    irq = (int) InB(nPnPReadPort);

    *pcIRQ          = irq;
    *pulBaseAddr    = iobase;
}

void SetupPnPCard(int CSN, int irq, int IOBASE)
{

    /* set CSN = totalCSN */
    OutB(0x06, PNP_ADDR_PORT);
    OutB(CSN, PNP_WR_PORT);

    /* Wake [CSN], set the card to config state */
    OutB(0x03, PNP_ADDR_PORT);
    OutB(CSN, PNP_WR_PORT);

    /* Set IRQ, IOBASE */
    /* IO PORT Address[15:8] */
    OutB(0x60, PNP_ADDR_PORT);    
    OutB((unsigned char) (IOBASE>>8)&0xFF, PNP_WR_PORT);

    /* IO PORT Address[7:0] */    
    OutB(0x61, PNP_ADDR_PORT);    
    OutB((unsigned char)IOBASE&0xFF, PNP_WR_PORT);
    
    /* IRQ Level */
    OutB(0x70, PNP_ADDR_PORT);    
    OutB((unsigned char) irq, PNP_WR_PORT);
    
    /* interrupt Type */
    OutB(0x71, PNP_ADDR_PORT);    
    OutB(0x02,PNP_WR_PORT);
    
    /* Active Logic Device */
    OutB(0x30, PNP_ADDR_PORT);    
    OutB(0x01, PNP_WR_PORT);
}

BOOL WL_PnpProbe(int nCardIndex, DWORD *pulBaseAddr, BYTE *pcIRQ)
{
    BOOL    bResult;
    int     nPnPRdPort;
    int     nSetWL;
    int     i;
    int     nTotalCSN;
    BYTE    buf[9];

    /* Return all cards to wait for key state */
    OutB(0x02, PNP_ADDR_PORT);
    NOPLOOP(100);
    OutB(0x02, PNP_WR_PORT);
    NOPLOOP(100);

    /* Set all card to sleep state */
    for (i = 0; i < 34; i++) {
        OutB(INIT_KEY_TABLE[i], PNP_ADDR_PORT);
        NOPLOOP(100);
    }

    /* Find ReadDataPort first */
    nPnPRdPort = Find_ReadDataPort();
    if (nPnPRdPort == 0) {
        printk("wlpnp.c: Maybe no WL2420 card on your PC or no free I/O port between\n");
        printk("         0x203 and 0x300.\n");
        return FALSE;
    }

    /* get total CSN and display all setup-card information */
    nTotalCSN = Find_TotalCSN(nPnPRdPort);

    bResult = FALSE;
    nSetWL = 0;
    for (i = 1; i <= nTotalCSN; i++) {
        GetResourceData(buf, nPnPRdPort, i);
        printk("wlpnp.c: Find an CSN ID :");
        TraceBuf(buf, sizeof(buf));
        if (*((DWORD*)buf) == WL_VENDOR_ID) {
            nSetWL++;
            
            if (nSetWL == nCardIndex) {
                WL_GetDeviceInfo(i, nPnPRdPort, pulBaseAddr, pcIRQ);
                bResult = TRUE;
                break;
            }
        }
    }
    
    /* Return all cards to wait for key state */
    OutB(0x02, PNP_ADDR_PORT);
    NOPLOOP(100);
    OutB(0x02, PNP_WR_PORT);
    NOPLOOP(100);

    return bResult;
}

void WL_PnpSetNonInitCard(DWORD ulBaseAddr, BYTE cIRQ)
{
    int     nPnPRdPort;
    int     nTotalCSN;
    int     i;
    BYTE    buf[9];

    /* Return all cards to wait for key state */
    OutB(0x02, PNP_ADDR_PORT);
    NOPLOOP(100);
    OutB(0x02, PNP_WR_PORT);
    NOPLOOP(100);

    /* Set all card to sleep state */
    for (i = 0; i < 34; i++) {
        OutB(INIT_KEY_TABLE[i], PNP_ADDR_PORT);
        NOPLOOP(100);
    }

    /* Find ReadDataPort first */
    nPnPRdPort = Find_ReadDataPort();
    if (nPnPRdPort == 0) {
        printk("wlpnp.c: Maybe no WL2420 card on your PC or no free I/O port between\n");
        printk("         0x203 and 0x300.\n");
        return;
    }

    /* get total CSN and display all setup-card information */
    nTotalCSN = Find_TotalCSN(nPnPRdPort);

    /* Finding NOTSET WL2420, and open the non-opend card */
    for (i = 0; i < DEV_NUM_MAX; i++) {
        GetSerialID(buf, nPnPRdPort);

        printk("wlpnp.c: Find an ID :");
        TraceBuf(buf, sizeof(buf));

        if (!CompareCheckSum(buf))
            break;

        if (*(DWORD*)buf == WL_VENDOR_ID){
            printk("wlpnp.c: Find a non-initialled WL24xx card, set IRQ=%d, IO=0x%3.3x\n",
                    (int) cIRQ, (int) ulBaseAddr);
            
            nTotalCSN++;
            SetupPnPCard(nTotalCSN, cIRQ, ulBaseAddr);
            break;
        } 
        else {
            /*  not wl2420 card */
            nTotalCSN++;
            /* set CSN to the isolation card */
            OutB(0x06, PNP_ADDR_PORT);
            NOPLOOP(100);
            OutB(nTotalCSN, PNP_WR_PORT);
            NOPLOOP(100);
        }
    }

    /* Return all cards to wait for key state */
    OutB(0x02, PNP_ADDR_PORT);
    NOPLOOP(100);
    OutB(0x02, PNP_WR_PORT);
    NOPLOOP(100);
}
